/*
 * Downloaded from lenkaspace.net
 * 
 * Paper about this project: http://lenkaspace.net/papers/iteratedPrisonersDilemmaNeuralNetworks.php
 * Email: contact@lenkaspace.net
 */

package net.lenkaspace.prisonersnorecognition;

import java.awt.Checkbox;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JTextField;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import net.lenkaspace.prisonersnorecognition.reports.ReportController;

public class ControlsPanel extends JPanel {

	private static final String REPORT_FITNESS = "Fitness";
	private static final String REPORT_COOPERATIONS = "Cooperations";
	private static final String REPORT_MUT_COOPERATIONS = "Mutual cooperations";
	private static final String REPORT_WEIGHTS = "Weights";
	private static final String REPORT_COOPERATIVE_STRATEGIES = "Cooperative strategies";
	private static final String REPORT_DEFECTIVE_STRATEGIES = "Defective strategies";
	private static final String REPORT_END_COOP_STRATEGIES = "End coop. strategies";
	private static final String REPORT_END_DEF_STRATEGIES = "End def. strategies";
	private static final String REPORT_END_OF_GEN_DEF_STRATEGIES = "EOG def. strategies";
	private static final String REPORT_END_OF_GEN_COOP_STRATEGIES = "EOG coop. strategies";
	
	/*private static final String INJECT_RANDOM = "Substitute random";
	private static final String INJECT_BEST = "Substitute best";
	private static final String INJECT_WORST = "Substitute worst"; */
	
	public static final String MODEL_STRAT = "Strategy-encoding";
	public static final String MODEL_FOGEL_NN = "Fogel-inspired neural network";
	public static final String MODEL_AS_NN = "Action separating neural network";
	
	public static final String NN_H1_TRANS = "1 T-layer";
	public static final String NN_H1 = "1 hidden layer";
	public static final String NN_H1_TRANS_H2 = "1 T-layer, 1 hidden";
	public static final String NN_H1_H2 = "2 hidden layers";
	public static final String NN_H1_H2_H3 = "3 hiddden layers";

	public Checkbox sm2chb;
	//--------------------
	private PrisonersNoRecognition mController;
	
	private JButton butTimeControl;
	private JButton	butSimControl;
	private JSlider sliderTimeSpeed;
	private JComboBox reportTypeCombo;
	//private JComboBox injectTypeCombo;
	public JComboBox agentTypeCombo;
	public JComboBox nnTypeCombo;
	
	public JTextField nnLayer1nodes;
	public JTextField nnLayer2nodes;
	public JTextField nnLayer3nodes;
	
	
	public JTextField recombProb;
	public JTextField mutProb;
	public JTextField population;
	public JTextField evolutions;
	public JTextField tournaments;
	public JTextField targetCoop;
	public JTextField parents;
	public JTextField mutVar;
	public JTextField demeSize;
	
	public JTextField parasitesNum;
	public JTextField parasitesW0;
	public JTextField parasitesW1;
	public JTextField parasitesW2;
	public JTextField parasitesW3;
	public JTextField parasitesW4;
	public JTextField parasitesVar;
	
	
	public Checkbox tournamentChB;
	public Checkbox generationOtherChB;
	public Checkbox generationStrategiesChB;
	public Checkbox generationCooperationsChB;
	
	
	public ControlsPanel(PrisonersNoRecognition controller_) {
		mController = controller_;
		
		this.setBounds(0, 0, 380, Settings.WINDOW_H-40);
		//this.setBackground(new Color(255,255,0));
		this.setSize (380,Settings.WINDOW_H-40);
		this.setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		
		//------------------------ agent settings:
		JPanel agentSettingsPanel = this.createFlowLayoutJPanel(40);
		agentSettingsPanel.add(new JLabel("Agent model:"));	
		
		//----- agent model combo box:
        String[] modelTypes = { MODEL_STRAT, MODEL_FOGEL_NN, MODEL_AS_NN };
        agentTypeCombo = new JComboBox(modelTypes);
        agentTypeCombo.setSelectedIndex(0);
        agentTypeCombo.addActionListener (new ActionListener () {
            public void actionPerformed(ActionEvent e) {
            	onAgentTypeComboChanged();
            }
        });
        
        agentSettingsPanel.add(agentTypeCombo);
        
		this.add(agentSettingsPanel);
		
		//----- neural network settings:
		JPanel nnPanel = this.createFlowLayoutJPanel(90);
		nnPanel.add(new JLabel("Neural network:"));	
			
        String[] nnTypes = { NN_H1_TRANS, NN_H1, NN_H1_TRANS_H2, NN_H1_H2, NN_H1_H2_H3 };
        nnTypeCombo = new JComboBox(nnTypes);     
        nnTypeCombo.setSelectedIndex(0);
        nnTypeCombo.addActionListener (new ActionListener () {
            public void actionPerformed(ActionEvent e) {
                onNnTypeComboChanged();
            }
        });
        nnPanel.add(nnTypeCombo);
        nnPanel.add(new JLabel("         "));
        
        nnPanel.add(new JLabel("Nodes in 1st layer:"));	
        nnLayer1nodes = this.createTextField("4");
        nnLayer1nodes.setColumns(2);
        nnPanel.add(nnLayer1nodes);
        nnPanel.add(new JLabel("2nd layer:"));
        nnLayer2nodes = this.createTextField("4");
        nnLayer2nodes.setColumns(2);
        nnPanel.add(nnLayer2nodes);
        nnPanel.add(new JLabel("3rd layer:"));
        nnLayer3nodes = this.createTextField("4");
        nnLayer3nodes.setColumns(2);
        nnPanel.add(nnLayer3nodes);
        
        sm2chb = new Checkbox("Use strict Sigmoid Middle (SM2)", false);
        sm2chb.setBackground(getBackground());
		nnPanel.add(sm2chb);
		
		//nnPanel.add(new JLabel("    "));
        
        this.add(nnPanel);
		
       
		
		this.add(this.createDivider());
		this.add(this.createSpace(10));
		//------------------ simulation settings:
		JPanel simSettingsPanel = new JPanel();
		Dimension gridSize = new Dimension(this.getSize().width, 120);
		simSettingsPanel.setSize(gridSize);
		simSettingsPanel.setPreferredSize(gridSize);
		simSettingsPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
		GridLayout gridLayout = new GridLayout(0,4);
		gridLayout.setHgap(5);
		gridLayout.setVgap(5);
		simSettingsPanel.setLayout(gridLayout);
		
		simSettingsPanel.add(new JLabel("Evolutions:"));
		evolutions = this.createTextField("1");
		simSettingsPanel.add(evolutions);
		simSettingsPanel.add(new JLabel("Tourn's /gen:"));
		tournaments = this.createTextField("5");
		simSettingsPanel.add(tournaments);
		
		simSettingsPanel.add(new JLabel("Population:"));
		population = this.createTextField("10");
		simSettingsPanel.add(population);
		simSettingsPanel.add(new JLabel("Parents /gen:")); 
		parents = this.createTextField("20");
		simSettingsPanel.add(parents);
		
		simSettingsPanel.add(new JLabel("Deme size:"));
		demeSize = this.createTextField("10");
		simSettingsPanel.add(demeSize);
		simSettingsPanel.add(new JLabel("Trgt coop. %:"));
		targetCoop = this.createTextField("0.9");
		simSettingsPanel.add(targetCoop);
		
		simSettingsPanel.add(new JLabel("Recomb. prob:"));
		recombProb = this.createTextField("1");
		simSettingsPanel.add(recombProb);
		simSettingsPanel.add(new JLabel("Mutat. prob:"));
		mutProb = this.createTextField("0.1");
		simSettingsPanel.add(mutProb);
		
		simSettingsPanel.add(new JLabel("Mutat. var:"));
		mutVar = this.createTextField("0.1");
		simSettingsPanel.add(mutVar);
		
		this.add(simSettingsPanel);
		
		this.add(this.createSpace(20));
		this.add(this.createDivider());
		//---------------------- output settings:
		this.add(this.createSpace(10));
		JPanel outpSettingsPanel = new JPanel();
		gridSize = new Dimension (50,100);
		outpSettingsPanel.setSize(gridSize);
		outpSettingsPanel.setPreferredSize(gridSize);
		outpSettingsPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
		gridLayout = new GridLayout(0,2);
		gridLayout.setHgap(5);
		gridLayout.setVgap(5);
		outpSettingsPanel.setLayout(gridLayout);
		
		outpSettingsPanel.add(new JLabel("Displayed output:"));
		tournamentChB = new Checkbox("Tournament", false);
		tournamentChB.setBackground(getBackground());
		outpSettingsPanel.add(tournamentChB);
		
		outpSettingsPanel.add(new JLabel(""));
		generationStrategiesChB = new Checkbox("Generation strategies", false);
		generationStrategiesChB.setBackground(getBackground());
		outpSettingsPanel.add(generationStrategiesChB);
		
		outpSettingsPanel.add(new JLabel(""));
		generationOtherChB = new Checkbox("Generation other", false);
		generationOtherChB.setBackground(getBackground());
		outpSettingsPanel.add(generationOtherChB);
		
		outpSettingsPanel.add(new JLabel(""));
		generationCooperationsChB = new Checkbox("Generation cooperations", false);
		generationCooperationsChB.setBackground(getBackground());
		outpSettingsPanel.add(generationCooperationsChB);
		
		this.add(outpSettingsPanel);
		
		this.add(new JLabel(" "));
	
		this.add(this.createDivider());
		this.add(this.createSpace(10));
		
		//------------------- start and pause button:
		JPanel timeButtonsPanel = this.createFlowLayoutJPanel(30);
		
		butSimControl = new JButton();
		butSimControl.setText("START NEW SIMULATION");
		butSimControl.addActionListener(new ActionListener() {	public void actionPerformed(ActionEvent e) {
                onSimulationButtonClick();
        }	}); 
		butSimControl.setAlignmentX(Component.LEFT_ALIGNMENT);
		timeButtonsPanel.add(butSimControl);
		
		butTimeControl = new JButton();
        butTimeControl.setText(">");
        butTimeControl.addActionListener(new ActionListener() {	public void actionPerformed(ActionEvent e) {
                onTimeButtonClick();
        }	}); 
        timeButtonsPanel.add(butTimeControl);
       
		
		this.add(timeButtonsPanel);
		
		
		
		//---------------------- time settings:
		
		JPanel timeSettingsPanel = this.createFlowLayoutJPanel(60);
		timeSettingsPanel.add(new JLabel("Time speed:"));	
		
		sliderTimeSpeed = new JSlider(0, 100, 0);
		sliderTimeSpeed.setPreferredSize(new Dimension (280,40));
		sliderTimeSpeed.setMinorTickSpacing(25);
		sliderTimeSpeed.setMajorTickSpacing(25);
		sliderTimeSpeed.setPaintTicks(true);
		sliderTimeSpeed.setPaintLabels(true);
		sliderTimeSpeed.addChangeListener( new ChangeListener() {
        	public void stateChanged(ChangeEvent e) {
        		onTimeSpeedChange();
        	}
        }); 
		timeSettingsPanel.add(sliderTimeSpeed);
		
		this.add(timeSettingsPanel);
		
		this.add(new JLabel(" "));
		this.add(this.createDivider());
		this.add(this.createSpace(10));
		
		 setTimeControlsEnabled(false);
		
		 
		//------------------ inject parasites:
		/*JPanel parasitesPanel = this.createFlowLayoutJPanel(30);
		parasitesPanel.add(new JLabel("Inject parasites:")); 
		parasitesNum = this.createTextField("5");
		parasitesPanel.add(parasitesNum);
		this.add(parasitesPanel);
		this.add(this.createSpace());
		
		this.add(new JLabel("Weight means:"));
		this.add(this.createSpace());
		
		JPanel parasitesWeightsPanel = new JPanel();
		gridSize = new Dimension (50,20);
		parasitesWeightsPanel.setSize(gridSize);
		parasitesWeightsPanel.setPreferredSize(gridSize);
		parasitesWeightsPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
		gridLayout = new GridLayout(0,10);
		gridLayout.setHgap(0);
		gridLayout.setVgap(10);
		parasitesWeightsPanel.setLayout(gridLayout);
		
		parasitesWeightsPanel.add(new JLabel("   W0:"));
		parasitesW0 = this.createTextField("0.1");
		parasitesWeightsPanel.add(parasitesW0);
		parasitesWeightsPanel.add(new JLabel("   W1:"));
		parasitesW1 = this.createTextField("0.1");
		parasitesWeightsPanel.add(parasitesW1);
		parasitesWeightsPanel.add(new JLabel("   W2:"));
		parasitesW2 = this.createTextField("0.1");
		parasitesWeightsPanel.add(parasitesW2);
		parasitesWeightsPanel.add(new JLabel("   W3:"));
		parasitesW3 = this.createTextField("0.1");
		parasitesWeightsPanel.add(parasitesW3);
		parasitesWeightsPanel.add(new JLabel("   W4:"));
		parasitesW4 = this.createTextField("0.1");
		parasitesWeightsPanel.add(parasitesW4);
		
		this.add(parasitesWeightsPanel);
		this.add(this.createSpace());
		//inject variance:
		JPanel parasitesMeanPanel = this.createFlowLayoutJPanel(24);
		parasitesMeanPanel.add(new JLabel("Weight variance:"));
		parasitesVar = this.createTextField("0.2");
		parasitesMeanPanel.add(parasitesVar);
		this.add(parasitesMeanPanel);
		this.add(this.createSpace());
		
		JPanel parasitesInjectPanel = this.createFlowLayoutJPanel(50);
		//inject type combo:
		String[] reportTypes = { INJECT_RANDOM, INJECT_BEST, INJECT_WORST};
        injectTypeCombo = new JComboBox(reportTypes);
        injectTypeCombo.setSelectedIndex(0);
        parasitesInjectPanel.add(injectTypeCombo);
        //inject button:
		JButton injectButton = new JButton();
		injectButton.setText("INFECT NEW GENERATION");
		injectButton.addActionListener(new ActionListener() {	public void actionPerformed(ActionEvent e) {
                onInjectParasitesClick();
        }	}); 
		injectButton.setAlignmentX(Component.LEFT_ALIGNMENT);
		parasitesInjectPanel.add(injectButton);
		
		this.add(parasitesInjectPanel);
		this.add(this.createDivider());
		this.add(this.createSpace()); */
		
		 
		//-------------------- reports:
		JPanel reportsPanel = this.createFlowLayoutJPanel(30);
		reportsPanel.add(new JLabel("Report:"));
		
		//----- type combo box:
        String[] injectTypes = { REPORT_FITNESS, REPORT_COOPERATIONS, REPORT_COOPERATIVE_STRATEGIES, REPORT_DEFECTIVE_STRATEGIES, REPORT_END_COOP_STRATEGIES, REPORT_END_DEF_STRATEGIES };
        reportTypeCombo = new JComboBox(injectTypes);
        reportTypeCombo.setSelectedIndex(0);
        reportsPanel.add(reportTypeCombo);
        
        //--- OPEN button
        JButton getReportButton = new JButton();
        getReportButton.setText("Open");
        getReportButton.addActionListener(new ActionListener() {	public void actionPerformed(ActionEvent e) {
                onOpenReportClick();
        }	});
        reportsPanel.add(getReportButton); 
		
        this.add(reportsPanel);
		/*
		this.add(new JLabel("Simulation type:"));
		
		JRadioButton radioFitnessYes = new JRadioButton("With fitness function (standard)");
		radioFitnessYes.setSelected(true);
		this.add(radioFitnessYes);
		
		JRadioButton radioFitnessNo = new JRadioButton("Without fitness function");
		this.add(radioFitnessNo);
		
		ButtonGroup radGroupSimulationType = new ButtonGroup();
		radGroupSimulationType.add(radioFitnessYes);
		radGroupSimulationType.add(radioFitnessNo);
		
		radioFitnessYes.addActionListener(new ActionListener() {	public void actionPerformed (ActionEvent e) {
				setSimulationType (Constants.SIMTYPE_FITNESS);
		}	});
		radioFitnessNo.addActionListener(new ActionListener() {	public void actionPerformed (ActionEvent e) {
			setSimulationType (Constants.SIMTYPE_NOFITNESS);
		}	});
		
		
	    
		
        */
        
        
        //perform input check:
        this.onAgentTypeComboChanged();
        
	}
	
	
	


	//=========================================================== ACTIONS
	
	//--------------------------------------- agent model type selection:
	
	private void onAgentTypeComboChanged() {
		String selectedNetworkType = (String) this.agentTypeCombo.getSelectedItem();
		if (selectedNetworkType == MODEL_STRAT) {
			setNNcontrolsEnabled(false);
		} else {
			//remove and add items to NN type based on NN model:
			nnTypeCombo.removeAllItems();
			if (selectedNetworkType == MODEL_FOGEL_NN) {	
				nnTypeCombo.addItem(NN_H1);
				nnTypeCombo.addItem(NN_H1_H2);
				nnTypeCombo.addItem(NN_H1_H2_H3);
			} else {
				nnTypeCombo.addItem(NN_H1_TRANS);
				nnTypeCombo.addItem(NN_H1);
				nnTypeCombo.addItem(NN_H1_TRANS_H2);
				nnTypeCombo.addItem(NN_H1_H2);
				nnTypeCombo.addItem(NN_H1_H2_H3);
			}
			setNNcontrolsEnabled(true);
		}
	}
	
	private void setNNcontrolsEnabled (boolean enabled) {
		nnTypeCombo.setEnabled(enabled);
		nnLayer1nodes.setEnabled(enabled);
		nnLayer2nodes.setEnabled(enabled);
		nnLayer3nodes.setEnabled(enabled);
		mutVar.setEnabled(enabled);
		sm2chb.setEnabled(enabled);
		mutProb.setEnabled(!enabled);
		if (enabled) {
			//perform additional check of the selected network type:
			this.onNnTypeComboChanged();
		}
	}
	
	private void onNnTypeComboChanged () {
		String selectedNetworkType = (String) this.nnTypeCombo.getSelectedItem();
		if (selectedNetworkType == NN_H1_TRANS || selectedNetworkType == NN_H1_TRANS_H2) {
			nnLayer1nodes.setText("4");
			nnLayer1nodes.setEnabled(false);
		} else {
			nnLayer1nodes.setEnabled(true);
		}
	}
	
	
	//--------------------------------------- parasites controls:
	private void onInjectParasitesClick() {
	
	}
	
	//-------------------------------------- report controls
	private void onOpenReportClick() {
		String selectedType = (String) this.reportTypeCombo.getSelectedItem();
		if (selectedType.equals(REPORT_FITNESS)) {
			ReportController.getSingletonObject().showReport(ReportController.REPORT_FITNESS);
		} else if (selectedType.equals(REPORT_COOPERATIONS)) {
			ReportController.getSingletonObject().showReport(ReportController.REPORT_COOPERATIONS);
		} else if (selectedType.equals(REPORT_MUT_COOPERATIONS)) {
			ReportController.getSingletonObject().showReport(ReportController.REPORT_MUT_COOPERATIONS);
		} else if (selectedType.equals(REPORT_WEIGHTS)) {
			ReportController.getSingletonObject().showReport(ReportController.REPORT_WEIGHTS);
		} else if (selectedType.equals(REPORT_COOPERATIVE_STRATEGIES)) {
			ReportController.getSingletonObject().showReport(ReportController.REPORT_COOPERATIVE_STRATEGIES);
		} else if (selectedType.equals(REPORT_DEFECTIVE_STRATEGIES)) {
			ReportController.getSingletonObject().showReport(ReportController.REPORT_DEFECTIVE_STRATEGIES);
		} else if (selectedType.equals(REPORT_END_COOP_STRATEGIES)) {
			ReportController.getSingletonObject().showReport(ReportController.REPORT_END_COOP_STRATEGIES);
		} else if (selectedType.equals(REPORT_END_DEF_STRATEGIES)) {
			ReportController.getSingletonObject().showReport(ReportController.REPORT_END_DEF_STRATEGIES);
		} else if (selectedType.equals(REPORT_END_OF_GEN_COOP_STRATEGIES)) {
			ReportController.getSingletonObject().showReport(ReportController.REPORT_END_OF_GEN_COOP_STRATEGIES);
		} else if (selectedType.equals(REPORT_END_OF_GEN_DEF_STRATEGIES)) {
			ReportController.getSingletonObject().showReport(ReportController.REPORT_END_OF_GEN_DEF_STRATEGIES);
		}
	}	

	//------------------------------------ simulation controls
	
	
	private void onSimulationButtonClick() {
		Settings settings = Settings.getSingletonObject();	
		mController.initSimulation();
		setTimeControlsEnabled(true);
		if (!settings.getTimeRunning()) {
			onTimeButtonClick();
		}
	}
	
	//-------------------------------------- time controls
	private void onTimeButtonClick() {
		Settings settings = Settings.getSingletonObject();
		
		if (settings.getTimeRunning()) {
			//---- pause clicked:
			mController.stopTime();
			butTimeControl.setText(">");
			settings.setTimeRunning(false);
			
		} else {
			//----- play clicked:	
			mController.startTime((double)sliderTimeSpeed.getValue() / 100);
			butTimeControl.setText("||");
			settings.setTimeRunning(true);
		}
		
	}
	
	private void onTimeSpeedChange() {
		if (Settings.getSingletonObject().getTimeRunning()) {
			mController.startTime((double)sliderTimeSpeed.getValue()/100);
		}
	}
	
	private void setTimeControlsEnabled(boolean enabled_) {
		sliderTimeSpeed.setEnabled(enabled_);
		butTimeControl.setEnabled(enabled_);
	}
	
	//======================================== COMPONENTS
	
	private JPanel createFlowLayoutJPanel(int height) {
		JPanel panel = new JPanel();
		Dimension panelSize = new Dimension (this.getSize().width, height);
		panel.setSize(panelSize);
		panel.setPreferredSize(panelSize);
		panel.setAlignmentX(Component.LEFT_ALIGNMENT);
		panel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
		return panel;
	}
	
	private JPanel createDivider() {
		JPanel panel = new JPanel();
		Dimension panelSize = new Dimension (this.getSize().width, 1);
		panel.setSize(panelSize);
		panel.setPreferredSize(panelSize);
		panel.setAlignmentX(Component.LEFT_ALIGNMENT - 10);
		panel.setLayout(new FlowLayout(FlowLayout.LEFT, 10, 10));
		panel.setBackground(Color.LIGHT_GRAY);
		return panel;
	}
	
	private JPanel createSpace(int height) {
		JPanel panel = new JPanel();
		Dimension panelSize = new Dimension (this.getSize().width, height);
		panel.setSize(panelSize);
		panel.setPreferredSize(panelSize);
		panel.setAlignmentX(Component.LEFT_ALIGNMENT - 10);
		panel.setLayout(new FlowLayout(FlowLayout.LEFT, 10, 10));
		return panel;
	}
	
	private JTextField createTextField(String defaultVal) {
		JTextField textField = new JTextField(5);
		textField.setAlignmentX(Component.LEFT_ALIGNMENT);
		textField.setPreferredSize(new Dimension(50,30));
		textField.setText(defaultVal);
		return textField;
	}
	
	
	
}

