/*
 * Downloaded from lenkaspace.net
 * 
 * Paper about this project: http://lenkaspace.net/papers/iteratedPrisonersDilemmaNeuralNetworks.php
 * Email: contact@lenkaspace.net
 */

package net.lenkaspace.prisonersnorecognition;

	import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JApplet;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.Timer;
import javax.swing.UIManager;

import net.lenkaspace.prisonersnorecognition.model.Agent;
import net.lenkaspace.prisonersnorecognition.model.World;
import net.lenkaspace.prisonersnorecognition.reports.ReportController;

	interface Constants
	{
	    public static final int SIMTYPE_FITNESS = 0;
	    public static final int SIMTYPE_NOFITNESS = 1;
	}



	public class PrisonersNoRecognition extends JApplet implements ActionListener {
		
		private OutputPanel outputPanel;
		

		private Timer timer;
		private World world;
		private ControlsPanel controlsPanel;
		
		private JPanel mainPanel;
		
		public void init() {
	        /* Turn off metal's use of bold fonts */
	        UIManager.put("swing.boldMetal", Boolean.FALSE);
	        this.getContentPane().setBounds(0, 0, Settings.WINDOW_W, Settings.WINDOW_H);
	        this.setSize(Settings.WINDOW_W, Settings.WINDOW_H);
	        //--------------- create GUI in a separate thread:
	        try {
	            javax.swing.SwingUtilities.invokeAndWait(new Runnable() {
	                public void run() {
	                	initGUI();             	
	                }
	            });
	        } catch (Exception e) {
	            System.err.println("initGUI didn't successfully complete");
	        }
	       
	        
	       
	    }
	 
	    public void initGUI() {
	    	setLayout(new FlowLayout(FlowLayout.LEFT, 10, 10));        
	    	mainPanel = new JPanel();
	        
	        //---- create output panel on the left:
	        outputPanel = new OutputPanel(this);
	        mainPanel.add(outputPanel);
	        add("West", mainPanel);
	        
	        mainPanel.add(new JLabel(""));
	        
	        //--- create controls panel on the right:
	        controlsPanel = new ControlsPanel(this);
	        mainPanel.add(controlsPanel);
	       // add("West", p);
	        
	          
	    }
	    
	    public void initSimulation() {
	    	
	    	if (this.checkUserInput()) {
	    		//set agent constants:
	    		Agent.USING_NN = !(controlsPanel.agentTypeCombo.getSelectedItem() == ControlsPanel.MODEL_STRAT);
	    		Agent.USING_FOGEL_INPUTS = (controlsPanel.agentTypeCombo.getSelectedItem() == ControlsPanel.MODEL_FOGEL_NN);
	    		Agent.USING_SM2 = controlsPanel.sm2chb.getState();
	    		String selectedNNType = (String) controlsPanel.nnTypeCombo.getSelectedItem();
	    		if (selectedNNType == ControlsPanel.NN_H1) {
	    			Agent.USING_NN_TYPE = Agent.NN_H1;
	    		} else if (selectedNNType == ControlsPanel.NN_H1_H2) {
	    			Agent.USING_NN_TYPE = Agent.NN_H1_H2;
	    		} else if (selectedNNType == ControlsPanel.NN_H1_H2_H3) {
	    			Agent.USING_NN_TYPE = Agent.NN_H1_H2_H3;
	    		} else if (selectedNNType == ControlsPanel.NN_H1_TRANS_H2) {
	    			Agent.USING_NN_TYPE = Agent.NN_H1_TRANS_H2;
	    		} else if (selectedNNType == ControlsPanel.NN_H1_TRANS) {
	    			Agent.USING_NN_TYPE = Agent.NN_H1_TRANS;
	    		}
	    		
	    		
	    		Agent.H1_NODES = Integer.valueOf(controlsPanel.nnLayer1nodes.getText());
	    		Agent.H2_NODES = Integer.valueOf(controlsPanel.nnLayer2nodes.getText());
	    		Agent.H3_NODES = Integer.valueOf(controlsPanel.nnLayer3nodes.getText());
		    	//set constants according to the controls panel input:
		    	World.OUTPUT_GENERATION_COOPERATIONS = controlsPanel.generationCooperationsChB.getState();
		    	World.OUTPUT_GENERATION_OTHER = controlsPanel.generationOtherChB.getState();
		    	World.OUTPUT_GENERATION_STRATEGIES = controlsPanel.generationStrategiesChB.getState();
		    	World.OUTPUT_TOURNAMENT = controlsPanel.tournamentChB.getState();
		    	
		    	World.NUM_AGENTS = Integer.valueOf(controlsPanel.population.getText());
		    	World.NUM_REPLICATED_AGENTS = Integer.valueOf(controlsPanel.parents.getText());
		    	World.NUM_EVOLUTIONS = Integer.valueOf(controlsPanel.evolutions.getText());
		    	World.TOURNAMENTS_PER_GENERATION = Integer.valueOf(controlsPanel.tournaments.getText());
		    	World.TARGET_COOPERATION_PERCENT = Double.valueOf(controlsPanel.targetCoop.getText());
		    	World.DEME_SIZE = Integer.valueOf(controlsPanel.demeSize.getText());
		    	
		    	Agent.RECOMBINATION_PROBABILITY = Double.valueOf(controlsPanel.recombProb.getText());
		    	Agent.MUTATION_PROBABILITY = Double.valueOf(controlsPanel.mutProb.getText());
		    	Agent.MUTATION_VARIANCE = Double.valueOf(controlsPanel.mutVar.getText());
		    	
		    	world = new World(this);
		    	ReportController.getSingletonObject().reset();
		    	this.outputPanel.resetOutputs();
		    	//------ tell settings that the init is done:
		        Settings.getSingletonObject().setInitDone(true);   
	    	}
	    }
	    
	    
	    private boolean checkUserInput() {
	    	if (Integer.valueOf(controlsPanel.population.getText()) <=0 || Integer.valueOf(controlsPanel.population.getText())%2 != 0) {
	    		JOptionPane.showMessageDialog(mainPanel, "Population must be > 0 and even", "", JOptionPane.ERROR_MESSAGE);
	    		return false;
	    	}	
	    	/*if (Integer.valueOf(controlsPanel.parents.getText()) > Integer.valueOf(controlsPanel.population.getText()) ) {
	    		JOptionPane.showMessageDialog(mainPanel, "Number of parents must <= population number", "", JOptionPane.ERROR_MESSAGE);
	    		return false;
	    	}*/
	    	if (Integer.valueOf(controlsPanel.evolutions.getText()) <= 0) {
	    		JOptionPane.showMessageDialog(mainPanel, "Number of evolutions must be at least 1", "", JOptionPane.ERROR_MESSAGE);
	    		return false;
	    	}
	    	if (Integer.valueOf(controlsPanel.tournaments.getText()) <= 0  || Integer.valueOf(controlsPanel.tournaments.getText()) > Integer.valueOf(controlsPanel.population.getText())/2) {
	    		JOptionPane.showMessageDialog(mainPanel, "Number of tournaments must be >=1 and <= population/2", "", JOptionPane.ERROR_MESSAGE);
	    		return false;
	    	}
	    	
	    	/*if (Double.valueOf(controlsPanel.targetCoop.getText()) < 0 || Double.valueOf(controlsPanel.targetCoop.getText()) > 1) {
	    		JOptionPane.showMessageDialog(mainPanel, "Target cooperation must be >= 0 and <= 1", "", JOptionPane.ERROR_MESSAGE);
	    		return false;
	    	}*/
	    	
	    	if (Double.valueOf(controlsPanel.recombProb.getText()) < 0 || Double.valueOf(controlsPanel.recombProb.getText()) > 1) {
	    		JOptionPane.showMessageDialog(mainPanel, "Recombination probability must be >= 0 and <= 1", "", JOptionPane.ERROR_MESSAGE);
	    		return false;
	    	}
	    	if (Double.valueOf(controlsPanel.mutProb.getText()) < 0 || Double.valueOf(controlsPanel.mutProb.getText()) > 1) {
	    		JOptionPane.showMessageDialog(mainPanel, "Mutation probability must be >= 0 and <= 1", "", JOptionPane.ERROR_MESSAGE);
	    		return false;
	    	}
	    	/*
	    	if (Double.valueOf(controlsPanel.mutVar.getText()) < 0 || Double.valueOf(controlsPanel.mutVar.getText()) > 1) {
	    		JOptionPane.showMessageDialog(mainPanel, "Mutation variance must be >= 0 and <= 1", "", JOptionPane.ERROR_MESSAGE);
	    		return false;
	    	} */
	    	if (Integer.valueOf(controlsPanel.demeSize.getText()) <= 1 || Integer.valueOf(controlsPanel.demeSize.getText()) %2 != 0) {
	    		JOptionPane.showMessageDialog(mainPanel, "Deme size must be >=2 and even", "", JOptionPane.ERROR_MESSAGE);
	    		return false;
	    	}
			return true;
		}

		//=========================================== MAIN LOOP
	    
	    public void actionPerformed(ActionEvent e) {
	    	// check if init has been done, if no do not continue:
	    	Settings settings = Settings.getSingletonObject();
	    	if (!settings.getInitDone()) {
	            return;
	        }
	    	
	    	//------------------ main loop:
	    	world.update();
			
		}
	    
	    //============================================= FUNCTIONALITY
	    
	    /*
	     * startTime
	     * 		- starts timer at a certain intervalLength
	     * 
	     * PARAMS:	intervalLength: [0..1] where 0 = slowest
	     * 
	     */
	    
	    public void startTime(double intervalLength) {
	    	if (Settings.getSingletonObject().getTimeRunning()) {
	    		timer.stop();
	    	}
	    	int maxInterval = 100;
	    	timer = new Timer((int) (maxInterval - maxInterval*intervalLength ), this); //time interval 5-30
	    	timer.setInitialDelay(0);
	    	timer.start();
	    }
	    
	    public void stopTime() {
	    	timer.stop();
	    }	
	    
	    //=================================================================== GETTTERS / SETTERS
	    
	    public OutputPanel getOutputPanel() {
			return outputPanel;
		}
}
