/*
 * Downloaded from lenkaspace.net
 * 
 * Paper about this project: http://lenkaspace.net/papers/iteratedPrisonersDilemmaNeuralNetworks.php
 * Email: contact@lenkaspace.net
 */

package net.lenkaspace.prisonersnorecognition.model;

import java.util.ArrayList;


import net.lenkaspace.prisonersnorecognition.Maths;
import net.lenkaspace.prisonersnorecognition.reports.ReportController;


/*
 * Agent class
 * - translates a genotype into an action, engages with other agents
 */
public class Agent extends BaseModel implements Comparable<Agent> {
	
	public static final int NN_H1_TRANS = 0;
	public static final int NN_H1 = 1;
	public static final int NN_H1_TRANS_H2 = 2;
	public static final int NN_H1_H2 = 3;
	public static final int NN_H1_H2_H3 = 4;
	
	
	//private static final int GENOTYPE_LENGTH = 24;
	public static  int GENOTYPE_LENGTH = 2;
	public static  int INPUTS_LENGTH = 12;
	//public static  int STRATEGY_LENGTH = 2;
	public static double RECOMBINATION_PROBABILITY = 1;
	public static double MUTATION_PROBABILITY = 0.1;
	public static double MUTATION_VARIANCE = 0.1; //
	
	//public static final String[] phenotypeNames = {"W0", "W1", "W1A", "W1B", "W2", "W2A", "W2B", "W2C", "W2D"};
	//public static final String[] phenotypeNames = {"W0", "W1", "W2", "W3", "W4", "W5", "W6", "W7", "W8", "W9", "W10", "W11", "I0", "I1", "I2", "I3"};
	//public static final String[] phenotypeNames = {"W0", "W1", "W2", "W3", "W4", "W5", "W6", "W7", "W8", "W9", "W10", "W11", "I0", "I1", "I2", "I3"};
	
	public static final int ACTIVATION_STEP = 0;
	public static final int ACTIVATION_LINEAR = 1;
	public static final int ACTIVATION_SIGMOID = 2;
	
	public static boolean USING_NN = true;
	public static boolean USING_FOGEL_INPUTS = false;
	public static boolean USING_SM2 = false;
	
	public static int USING_NN_TYPE = Agent.NN_H1_H2;
	public static int H1_NODES = 20;
	public static int H2_NODES = 20;
	public static int H3_NODES = 8;
	
	private boolean playingTFT;
	
	private ArrayList<Integer> alreadyPlayedWithList;

	//------------
	private double[] genotype;
	

	private double[] inputs;
	private int payoff;
	private int roundsPlayed;
	private int agentNo;
	
	private String strategiesRecord;
	private double[][]inputCombinationsForNN;
	
	ArrayList <String> possibleScenarios;
	
	public enum Action {
	    COOPERATE, DEFECT     
	}
	
	public Agent(int id_, World world_)  {
		super(id_);
		
		inputCombinationsForNN = new double [16] [Agent.INPUTS_LENGTH];
		double[]opt0 = { 1,0, 1,0, 1,0, 1,0}; inputCombinationsForNN[0] = opt0;
		double[]opt1 = { 1,0, 1,0, 0,1, 1,0}; inputCombinationsForNN[1] = opt1;
		double[]opt2 = { 1,0, 1,0, 1,0, 0,1}; inputCombinationsForNN[2] = opt2;
		double[]opt3 = { 1,0, 1,0, 0,1, 0,1}; inputCombinationsForNN[3] = opt3;
		
		double[]opt4 = { 1,0, 0,1, 1,0, 1,0}; inputCombinationsForNN[4] = opt4;
		double[]opt5 = { 1,0, 0,1, 0,1, 1,0}; inputCombinationsForNN[5] = opt5;
		double[]opt6 = { 1,0, 0,1, 1,0, 0,1}; inputCombinationsForNN[6] = opt6;
		double[]opt7 = { 1,0, 0,1, 0,1, 0,1}; inputCombinationsForNN[7] = opt7;
		
		double[]opt8 = { 0,1, 1,0, 1,0, 1,0}; inputCombinationsForNN[8] = opt8;
		double[]opt9 = { 0,1, 1,0, 0,1, 1,0}; inputCombinationsForNN[9] = opt9;
		double[]opt10 = { 0,1, 1,0, 1,0, 0,1}; inputCombinationsForNN[10] = opt10;
		double[]opt11 = { 0,1, 1,0, 0,1, 0,1}; inputCombinationsForNN[11] = opt11;
		
		double[]opt12 = { 0,1, 0,1, 1,0, 1,0}; inputCombinationsForNN[12] = opt12;
		double[]opt13 = { 0,1, 0,1, 0,1, 1,0}; inputCombinationsForNN[13] = opt13;
		double[]opt14 = { 0,1, 0,1, 1,0, 0,1}; inputCombinationsForNN[14] = opt14;
		double[]opt15 = { 0,1, 0,1, 0,1, 0,1}; inputCombinationsForNN[15] = opt15;
		
		
		possibleScenarios = new ArrayList<String>();
		
		possibleScenarios.add("1111"); //RR 
		possibleScenarios.add("1101"); //RS 
		possibleScenarios.add("1110"); //RT 
		possibleScenarios.add("1100"); //RP
		
		possibleScenarios.add("1011"); //TR
		possibleScenarios.add("1001"); //TS
		possibleScenarios.add("1010"); //TT
		possibleScenarios.add("1000"); //TP
		
		possibleScenarios.add("0111"); //SR
		possibleScenarios.add("0101"); //SS
		possibleScenarios.add("0110"); //ST
		possibleScenarios.add("0100"); //SP
		
		possibleScenarios.add("0011"); //PR 
		possibleScenarios.add("0001"); //PS 
		possibleScenarios.add("0010"); //PT
		possibleScenarios.add("0000"); //PP 
			
		this.initState();
		this.genotype = new double [GENOTYPE_LENGTH];
		//create a random genotype:
		if (USING_NN) {
			//generate initial weights and inputs as double numbers, inputs will be later translated to 0s and1s
			if (!USING_FOGEL_INPUTS) {
				//numbers between 0 and 1
				for (int i=0; i<GENOTYPE_LENGTH; i++) {
					Double gene = Maths.getSingletonObject().getRandomDouble(true);
					this.genotype[i] = gene;
				}
			} else {
				//numbers between -1 and 1
				for (int i=0; i<GENOTYPE_LENGTH; i++) {
					//weigths
					Double gene = (double)Maths.getSingletonObject().getRandomInteger(-100, 100)/(double)100;
					this.genotype[i] = gene;
				}
			}
		} else {
			for (int i=0; i<GENOTYPE_LENGTH; i++) {				
				Double rand = Maths.getSingletonObject().getRandomDouble(true);
				if (rand > 0.5) {
					this.genotype[i] = 1;
				} else {
					this.genotype[i] = 0;
				}	
			}			
		}
		this.initGameState();
	}
	
/*	public Agent(int id_, World world_, Agent parent1, Agent parent2)  {
		super(id_);
		this.initState();
		this.genotype = new double [GENOTYPE_LENGTH];
		if (parent1.getFitness() > parent2.getFitness()) {
			for (int g=0; g<GENOTYPE_LENGTH; g++) {
				this.genotype[g] = parent2.getGenotype()[g];
			}
			this.changeGenotype(parent1);
		} else {
			for (int g=0; g<GENOTYPE_LENGTH; g++) {
				this.genotype[g] = parent1.getGenotype()[g];
			}
			this.changeGenotype(parent2);
		}
		this.createStartingInputs();
	} */
		
	public void initState() {
		if (USING_NN) {
			int INPUTS_LENGTH_IN_GENOTYPE = 0;
			if (!USING_FOGEL_INPUTS) {
				INPUTS_LENGTH = 8;
				INPUTS_LENGTH_IN_GENOTYPE = 4;
			} else {
				INPUTS_LENGTH = 4;
				INPUTS_LENGTH_IN_GENOTYPE = 4;
			}
			if (USING_NN_TYPE == Agent.NN_H1_TRANS) {
				GENOTYPE_LENGTH = 21; //8+4 + 4(bias) +8/2 + 1 //I, HL1 with 4 nodes,translation, O
				//GENOTYPE_LENGTH = 15; //8+2 + 2(bias) + 8/2 + 1 //I, HL1 with 2 nodes,translation, O
			} else if (USING_NN_TYPE == Agent.NN_H1_TRANS_H2) {
				GENOTYPE_LENGTH = INPUTS_LENGTH + H1_NODES + (H1_NODES+1)*H2_NODES + (H2_NODES+1)*1 + INPUTS_LENGTH_IN_GENOTYPE;
				//GENOTYPE_LENGTH = 41; //8+4(bias) + 5*4 + 5*1  + 8/2 //I, HL1 with 4 nodes, H2with4, translation, O
			} else if (USING_NN_TYPE == Agent.NN_H1) {
				GENOTYPE_LENGTH = (INPUTS_LENGTH + 1)*H1_NODES + (H1_NODES + 1)*1 + INPUTS_LENGTH_IN_GENOTYPE;
			} else if (USING_NN_TYPE == Agent.NN_H1_H2) {
				GENOTYPE_LENGTH = (INPUTS_LENGTH + 1)*H1_NODES + (H1_NODES+1)*H2_NODES + (H2_NODES + 1)*1 + INPUTS_LENGTH_IN_GENOTYPE;
			} else if (USING_NN_TYPE == Agent.NN_H1_H2_H3) {
				GENOTYPE_LENGTH = (INPUTS_LENGTH + 1)*H1_NODES + (H1_NODES+1)*H2_NODES + (H2_NODES+1)*H3_NODES + (H3_NODES + 1)*1 + INPUTS_LENGTH_IN_GENOTYPE;
			}		
		} else {
			GENOTYPE_LENGTH = 20; //4*4+4
			INPUTS_LENGTH = 4;		
		}
		this.payoff = 0;
		this.roundsPlayed = 0;
		this.inputs = new double[INPUTS_LENGTH];
		this.strategiesRecord = "";
		this.alreadyPlayedWithList = new ArrayList <Integer>();
		this.initGameState();
	}
	
	public void initGameState() {
		//create initial hypothetical inputs based on last N genes:
		if (this.genotype != null) {
			if (USING_NN) {
				if (!USING_FOGEL_INPUTS) {
					int counter = 0;
					for (int i=0; i<INPUTS_LENGTH/2; i++) {
						double geneVal = this.genotype[i + (GENOTYPE_LENGTH - INPUTS_LENGTH/2)];
						if (geneVal > 0.5) {
							this.inputs[counter++] = 1;
							this.inputs[counter++] = 0;
						} else {
							this.inputs[counter++] = 0;
							this.inputs[counter++] = 1;
						}
					}
				} else {
					for (int i=0; i<INPUTS_LENGTH; i++) {
						double geneVal = this.genotype[i + (GENOTYPE_LENGTH - INPUTS_LENGTH)];
						if (geneVal > 0) {
							this.inputs[i] = 1;
						} else {
							this.inputs[i] = -1;
						}
					}
				}
			}
			else {
				for (int i=0; i<INPUTS_LENGTH; i++) {
					inputs[i] = this.genotype[GENOTYPE_LENGTH - INPUTS_LENGTH+i];
				}
			}
		}
		this.playingTFT = true;
	}
	
	
	
	
	
	//======================================================================= PRISONERS DILEMMMA
	public Agent.Action doAction() {
		//compute cooperation probability:
		double cooperationProb = genotype[0];
		double randNum = Maths.getSingletonObject().getRandomDouble(true);
		
		if (USING_NN) {
			cooperationProb = this.calculateCooperationProbWithNN(this.inputs);	
		} else {
			String str = "";
			for (int i=0; i<INPUTS_LENGTH; i++) {
				str += (int)this.inputs[i];
			}
			int whatHappenedIndex = possibleScenarios.indexOf(str);
			cooperationProb = this.genotype[whatHappenedIndex];		
		}	
		// decide if C or D:
		if (randNum <= cooperationProb && cooperationProb != 0) {  
			return Agent.Action.COOPERATE;
		} else {
			return Agent.Action.DEFECT;
		}	
	}
	
	
	
	private double calculateCooperationProbWithNN(double[] inputLayer) {	
		double cooperationProb = 0;
		int weightsOffset = 0;
		int hiddenLayer1Size = H1_NODES; //set by UI
		double []hiddenLayer1 = new double[hiddenLayer1Size];
		
		
		//calculate hidden layer 1
		if (USING_NN_TYPE == Agent.NN_H1_TRANS || USING_NN_TYPE == Agent.NN_H1_TRANS_H2 ) {
			//=============== 'translation' hidden layer
			for (int i=0; i<hiddenLayer1Size; i++) {
				/*========= all actions from 1 timestep, 1 actor to 1 node (H1=4) ========= */
				double a = this.genotype[i*3] * inputLayer[i*2];
				double b = this.genotype[i*3+1]*inputLayer[i*2+1];
				double c = this.genotype[i*3+2]*1; //bias
				hiddenLayer1[i] = (double)1/2* (a+b) + (double)1/2*c; //max val: 1 
																		// values multiplied by 0.5, becayse max for (a+b) = 1
																		// max for bias = 1, so we effectively have as if 2 inputs 
				/*========= all actions from 1 timestep to 1 node (H1=2) ========= */
				/*for (int j=0; j<4; j++) {
					int whereGen = i*5 + j;
					int whereInp = i*4 + j;
					hiddenLayer1[i] += (double)1/3* this.genotype[whereGen] * inputLayer[whereInp];
				}
				hiddenLayer1[i]  += (double)1/3 * this.genotype[i*5 + 4]; //bias  */
				/*========= all actions from 1 timestep to 1 node (H1=1) ========= */
			/*	for (int j=0; j<8; j++) {
					int whereGen = i*9 + j;
					int whereInp = i*8 + j;
					hiddenLayer1[i] += (double)1/5* this.genotype[whereGen] * inputLayer[whereInp];
				}
				hiddenLayer1[i]  += (double)1/5 * this.genotype[i*9 + 8]; //bias */
			}
			//activate hidden layer 1:
			for (int i=0; i<hiddenLayer1Size; i++) {
				hiddenLayer1[i] = this.activate(hiddenLayer1[i], Agent.ACTIVATION_SIGMOID, 0.5); //middle is 0.5, had only 2 nodes attached to each 
			} 
			weightsOffset = INPUTS_LENGTH+hiddenLayer1Size; //inputs + number of biases  
		} else {
			//======== fully connected first layer:
			for (int i=0; i<hiddenLayer1Size; i++) {
				double weightOfConnection = 0;
				if (!USING_FOGEL_INPUTS) {
					weightOfConnection = (double)1/(INPUTS_LENGTH/2+1); // only INPUTS_LENGTH/2, because half of the inputs
																		// will have value 0
				} else {
					weightOfConnection = (double)1/(INPUTS_LENGTH+1);
				}
				for (int j=0; j<INPUTS_LENGTH; j++) {
					int where = i*(INPUTS_LENGTH+1) + j;	
					hiddenLayer1[i] +=  weightOfConnection * (this.genotype[where] * inputLayer[j]); //max val would be 4
				}
				int where = i*(INPUTS_LENGTH+1) + INPUTS_LENGTH;
				hiddenLayer1[i] += weightOfConnection * this.genotype[where]; // add bias			
			}
			double resultMiddle;
			if (!USING_FOGEL_INPUTS) {
				resultMiddle = 0.5 ;//(double)1/(INPUTS_LENGTH/2) * (double)5/6;
			} else {
				resultMiddle = 0.5; //(double)1/(INPUTS_LENGTH) * (double)5/6;
			}
			//activate hidden layer 1:
			for (int i=0; i<hiddenLayer1Size; i++) {
				hiddenLayer1[i] = this.activate(hiddenLayer1[i], Agent.ACTIVATION_SIGMOID, resultMiddle);
			}	
			weightsOffset = hiddenLayer1Size*(INPUTS_LENGTH+1);  
		}
		
		
		//============== second layer:	
		int hiddenLayer2Size = H2_NODES;
		double []hiddenLayer2 = new double[hiddenLayer2Size];
		
		if (USING_NN_TYPE == Agent.NN_H1_TRANS_H2 || USING_NN_TYPE >= Agent.NN_H1_H2 ) {	
			for (int i=0; i<hiddenLayer2Size; i++) {
				for (int j=0; j<hiddenLayer1Size; j++) {
					int where = weightsOffset + i*(hiddenLayer1Size+1) + j;
					hiddenLayer2[i] += (double)1/(hiddenLayer1Size+1) * (this.genotype[where] * hiddenLayer1[j]);
				}
				int where2 = weightsOffset + i*(hiddenLayer1Size+1) + hiddenLayer1Size;
				hiddenLayer2[i] += (double)1/(hiddenLayer1Size+1) * this.genotype[where2]; // add bias
			}
			double resultMiddle = 0.5;//(double)1/hiddenLayer1Size * (double)5/6;
			//activate hidden layer 2:
			for (int i=0; i<hiddenLayer2Size; i++) {
				hiddenLayer2[i] = this.activate(hiddenLayer2[i], Agent.ACTIVATION_SIGMOID, resultMiddle);
			}
			weightsOffset = weightsOffset + hiddenLayer2Size*(hiddenLayer1Size+1);
		}
		
		//============== third layer:	
		int hiddenLayer3Size = H3_NODES;
		double []hiddenLayer3 = new double[hiddenLayer3Size];
		
		if (USING_NN_TYPE >= Agent.NN_H1_H2_H3) {	
			for (int i=0; i<hiddenLayer3Size; i++) {
				for (int j=0; j<hiddenLayer2Size; j++) {
					int where = weightsOffset + i*(hiddenLayer2Size+1) + j;
					hiddenLayer3[i] += (double)1/(hiddenLayer2Size+1) * (this.genotype[where] * hiddenLayer2[j]);
				}
				int where2 = weightsOffset + i*(hiddenLayer2Size+1) + hiddenLayer2Size;
				hiddenLayer3[i] += (double)1/(hiddenLayer2Size+1) * this.genotype[where2]; // add bias
			}
			double resultMiddle = 0.5; //(double)1/hiddenLayer2Size * (double)5/6;
			//activate hidden layer 2:
			for (int i=0; i<hiddenLayer3Size; i++) {
				hiddenLayer3[i] = this.activate(hiddenLayer3[i], Agent.ACTIVATION_SIGMOID, resultMiddle);
			}
			weightsOffset = weightsOffset + hiddenLayer3Size*(hiddenLayer2Size+1);
		}
		
		//=========== output	
		int lastLayerSize = 0;
		double []lastLayer = null;
		if (USING_NN_TYPE == Agent.NN_H1_TRANS || USING_NN_TYPE == Agent.NN_H1) {
			lastLayerSize = hiddenLayer1Size;
			lastLayer = hiddenLayer1;
		} else if (USING_NN_TYPE < NN_H1_H2_H3){
			lastLayerSize = hiddenLayer2Size;
			lastLayer = hiddenLayer2;
		} else {
			lastLayerSize = hiddenLayer3Size;
			lastLayer = hiddenLayer3;
		}
		
		//=========== calculate output based on last layer:
		double output = 0;	
		if (USING_SM2) {
			lastLayerSize = 4;
		}
		double resultMiddle = (double)1/lastLayerSize * (double)5/6;
		for (int i=0; i<lastLayerSize; i++) {
			double a = this.genotype[weightsOffset+i]*(double)lastLayer[i];
			output += (double)1/(lastLayerSize+1) * a ; //maxVal:1
		}
		int where = weightsOffset + lastLayerSize;
		output += (double)1/ (lastLayerSize+1) * this.genotype[where]; //add bias */
	
		//activate output:
		output = activate(output, Agent.ACTIVATION_SIGMOID, resultMiddle);
		
		//decide if C or D, set cooperation probability to 1 or 0 depending on value of output:
		if (!USING_FOGEL_INPUTS) {
			cooperationProb = activate(output, Agent.ACTIVATION_STEP, 0.5);
		} else {
			cooperationProb = activate(output, Agent.ACTIVATION_STEP, 0.0); //all above 0 cooperate
		}
		return cooperationProb;
	}

	private double activate(double valueIn, int activationType, double middle) {
		double result = 0;
		if (activationType == Agent.ACTIVATION_LINEAR) {
			result = valueIn;
		} else if (activationType == Agent.ACTIVATION_STEP) {
			if (valueIn > middle) {
				result = 1;
			} else {
				result = 0;
			}
		} else if (activationType == Agent.ACTIVATION_SIGMOID) {
			double temp = (double)(12* (0.5 / middle));// will move middle of sigmoid curve to the middle specified as param of function
			if (USING_NN) {
				if (!USING_FOGEL_INPUTS) {
					result = (1/ (1 + Math.exp(-(-6 + temp*valueIn)))); //sigmoids have different middles but return val always from <0;1>
				} else {
					result = -1 + 2*(1/ (1 + Math.exp(-(-0 + 7*valueIn)))); //sigmoids have different middles but return val always from <-1;1>
					//result = -1 + 2*(1/ (1 + Math.exp(-(-7 + 14*valueIn))));
				}
			} else {
				result = (1/ (1 + Math.exp(-(-48 + 96*valueIn))));
			}
		}
		return result;
	}


	public void registerActions(Agent.Action opponentAction, Agent.Action selfAction) {	
		//push array of memory:
		if (USING_NN) {
			if (!USING_FOGEL_INPUTS) {
				//push self action
				inputs[6] = inputs[2];
				inputs[7] = inputs[3];
				//push other action
				inputs[4] = inputs[0];
				inputs[5] = inputs[1];			
				if (opponentAction == Agent.Action.COOPERATE) {
					inputs[0] = 1.0;
					inputs[1] = 0.0;
				} else {
					inputs[0] = 0.0;
					inputs[1] = 1.0;
				}		
				if (selfAction == Agent.Action.COOPERATE) {
					inputs[2] = 1.0;
					inputs[3] = 0.0;
				} else {
					inputs[2] = 0.0;
					inputs[3] = 1.0;
				}
			} else {
				//push self action
				inputs[3] = inputs[1];
				//push other action
				inputs[2] = inputs[0];	
				if (opponentAction == Agent.Action.COOPERATE) {
					inputs[0] = 1.0;
				} else {
					inputs[0] = -1.0;
				}			
				if (selfAction == Agent.Action.COOPERATE) {
					inputs[1] = 1.0;
				} else {
					inputs[1] = -1.0;
				}
			}
		} else {
			//push memory:
			this.inputs[2] = this.inputs[0];
			this.inputs[3] = this.inputs[1];
			if (opponentAction == Agent.Action.COOPERATE) {
				this.inputs[0] = 1;
			} else {
				this.inputs[0] = 0;
			}
			if (selfAction == Agent.Action.COOPERATE) {
				this.inputs[1] = 1;
			} else {
				this.inputs[1] = 0;
			}		
		}	
	}
	
	
	//========================================= GENOTYPE
	
	public void changeGenotype(Agent other) {
		//go through each gene of winner, copy it onto the looser with some probability and mutate with some probability:
		for (int g=0; g<GENOTYPE_LENGTH; g++ ) {
			//Double gene = this.genotype.[g);
			if (Maths.getSingletonObject().getRandomDouble(true) < RECOMBINATION_PROBABILITY) {
				this.genotype[g] = other.getGenotype()[g];
			}
			
			if (USING_NN) {
				/*if (g<GENOTYPE_LENGTH - INPUTS_LENGTH) {
					//mutate using gaussian distribution
					this.genotype[g] = Maths.getSingletonObject().getRandomGaussian(this.genotype[g], MUTATION_VARIANCE); 
					//make sure gene value is between 0 and 1:
					if (this.genotype[g] < 0) {
						this.genotype[g] = 0.0;
					} else if (this.genotype[g] > 1) {
						this.genotype[g] = 1.0;
					}
				} else {
					if (Maths.getSingletonObject().getRandomDouble(true) < MUTATION_PROBABILITY) {
						if (this.genotype[g] == 1) {
							this.genotype[g] = 0.0;
						} else {
							this.genotype[g] = 1.0;
						}
					}
				} */
				this.genotype[g] = Maths.getSingletonObject().getRandomGaussian(this.genotype[g], MUTATION_VARIANCE); 
				if (!USING_FOGEL_INPUTS) {
					//make sure gene value is between 0 and 1:
					if (this.genotype[g] < 0) {
						this.genotype[g] = 0.0;
					} else if (this.genotype[g] > 1) {
						this.genotype[g] = 1.0;
					}
				} else {
					//make sure gene value is between -1 and 1:
					if (this.genotype[g] < -1) {
						this.genotype[g] = -1.0;
					} else if (this.genotype[g] > 1) {
						this.genotype[g] = 1.0;
					}
				}
			} else {
				if (Maths.getSingletonObject().getRandomDouble(true) < MUTATION_PROBABILITY) {
					if (this.genotype[g] == 1) {
						this.genotype[g] = 0.0;
					} else {
						this.genotype[g] = 1.0;
					}
				}
			}
		}
	}
	
	public double[] getGenotypeAsStratFromNN() {
		//hard-code inputs and test action, put into genotype[16], to get the same output as agent.getGenotype() for evolution with strategies:
		double [] genotype = new double[ReportController.NUM_STRATEGIES];
		for (int i=0; i<ReportController.NUM_STRATEGIES; i++) {
			double randNum = Maths.getSingletonObject().getRandomDouble(true);
			double myWouldBeCoopProb = this.calculateCooperationProbWithNN(inputCombinationsForNN[i]);
			// decide if C or D:
			if (randNum <= myWouldBeCoopProb && myWouldBeCoopProb != 0) { //  //cooperationProb >0.5
				genotype[i] = 1;
			} else {
				genotype[i] = 0;
			}	
		}
		return genotype;
	}
	
	
	//===========================================================
	
	public void addPlayedWith(int agentNo) {
		if (!this.alreadyPlayedWithList.contains(agentNo)) {
			this.alreadyPlayedWithList.add(agentNo);
		}
	}
	
	public int compareTo(Agent other) {
		if (other.getFitness() < this.getFitness()) {
			return -1;
		} else if (other.getFitness() > this.getFitness()) {
			return 1;
		} else {
			//both same fitness, decide randomly:
			double randNum = Maths.getSingletonObject().getRandomDouble(true);
			if (randNum <= 0.5) {
				return 1;
			} else {
				return -1;
			}
		}
		
	}

	
	//========================================================== GETTERS / SETTERS
	public double[] getGenotype() {
		return genotype;
	}
	
	public void setPayoff(int payoff) {
		this.payoff = payoff;
		this.roundsPlayed++;
	}

	public int getPayoff() {
		return payoff;
	}
	
	public double getFitness() {
		return (double)this.payoff / (double)this.roundsPlayed;
	}
	
	
	

	public String translateDna(boolean includeAgentId) {
		
		String returnStr = "";
		if (includeAgentId) {
			returnStr += "("+this.id+") ";
		}
		if (USING_NN) {
			/*for (int i=0; i<GENOTYPE_LENGTH- INPUTS_LENGTH/2; i++) {
				//returnStr += Agent.phenotypeNames[i] +"="+ this.genotype[i] + " ";
				returnStr += "W"+i +"="+ this.genotype[i] + " ";
			}*/
			double[] wouldBeGenot = this.getGenotypeAsStratFromNN();
			for (int i=0; i<ReportController.NUM_STRATEGIES; i++) {
				
				returnStr += "" + (int)wouldBeGenot[i] + possibleScenarios.get(i) + " ";
				
			}
		} else {
			
			for (int i=0; i<GENOTYPE_LENGTH; i++) {
				if (i<GENOTYPE_LENGTH - INPUTS_LENGTH) {
					returnStr += "" + (int)this.genotype[i] + possibleScenarios.get(i) + " ";
				} else {
					returnStr += (int)this.genotype[i];
				}
			}
		}
		return returnStr;
	
	}
	
	public String translateInputs() {
		String returnStr = "";
		if (USING_NN) {
			/*for (int i=0; i<INPUTS_LENGTH; i++) {
				returnStr += Agent.phenotypeNames[i + (GENOTYPE_LENGTH - INPUTS_LENGTH)] +"="+ this.inputs[i] + " ";
			}*/
		} else {
			
			for (int i=0; i<INPUTS_LENGTH; i++) {
				returnStr += (int)this.inputs[i];
			}
		}
		return returnStr;
	
	}
	
	public double[] getInputs() {
		return this.inputs;
	}
	
	public void setInputs(double[]newInputs) {
		this.inputs = newInputs;
	}



	public void setStrategiesRecord(String strategiesRecord) {
		this.strategiesRecord = strategiesRecord;
	}
	
	public void addToStrategiesRecord(String additionStr) {
		this.strategiesRecord += additionStr;
	}


	public String getStrategiesRecord() {
		return strategiesRecord;
	}


	public void setAgentNo(int agentNo) {
		this.agentNo = agentNo;
	}


	public int getAgentNo() {
		return agentNo;
	}
	
	public String getStrategiesSummary() {
		return "AGENT " + this.agentNo + "(fitness: "+this.getFitness()+"): " + this.strategiesRecord;
	}

	public boolean alreadyPlayedWith(Integer nextInt) {
		if (this.alreadyPlayedWithList.contains(nextInt)) {
			return true;
		}
		return false;
	}

	public void setPlayingTFT(boolean playingTFT) {
		this.playingTFT = playingTFT;
	}

	public boolean isPlayingTFT() {
		return playingTFT;
	}
	
	
	
	
	
}
