/*
 * Downloaded from lenkaspace.net
 * 
 * Paper about this project: http://lenkaspace.net/papers/iteratedPrisonersDilemmaNeuralNetworks.php
 * Email: contact@lenkaspace.net
 */

package net.lenkaspace.prisonersnorecognition.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import net.lenkaspace.prisonersnorecognition.Maths;
import net.lenkaspace.prisonersnorecognition.PrisonersNoRecognition;
import net.lenkaspace.prisonersnorecognition.reports.ReportController;

/*
 * World class
 * - contains agents and implements the genetic algorithm
 */
public class World  {
	//---- constats
	public static int NUM_AGENTS = 200; //must be even number
	public static int TOURNAMENTS_PER_GENERATION = 10; 
	public static int NUM_REPLICATED_AGENTS = 100; //cannot be more than NUM_AGENTS
	//private static int TOURNAMENT_ROUNDS = 3;
	
	public static boolean OUTPUT_TOURNAMENT = false;
	public static boolean OUTPUT_GENERATION_STRATEGIES = false;
	public static boolean OUTPUT_GENERATION_COOPERATIONS = false;
	public static boolean OUTPUT_GENERATION_OTHER = false;
	
	public static int NUM_EVOLUTIONS = 1000;
	public static final int MAX_GENERATIONS = 10000;
	
	public static double TARGET_COOPERATION_PERCENT = 0.85;
	
	public static int DEME_SIZE = 200;
	
//	public static String generationStrategiesStr;
		
	//----- objects
	private ArrayList <Agent> agentList;
	private PrisonersNoRecognition owner;
	//--- update loop
	private Object[] tournamentPairs;	
	
	private int tournamentNo = -1;
	private int generation = 0;
	private int evolution = 0;
	
	
	private int counter = 0;
	
	
	//stats
	private int cooperationsCount;
	private int mutCooperationsCount;
	private int actionsCount;
	
	private int totalGenerations;
	private int minGenerations;
	private int maxGenerations;
	
	private ArrayList<String> endOfEvolutionStrategies; 
	private int minUniqueStrategies;
	private int maxUniqueStrategies;
	private int totalUniqueStrategies;
	
	private double minUniqueWinningStrategies;
	private double maxUniqueWinningStrategies;
	private double totalUniqueWinningStrategies;
	
	private double totalWinningFitness;
	private double minWinningFitness;
	private double maxWinningFitness;
	
	private double totalAverageFitness;
	private double minAverageFitness;
	private double maxAverageFitness;
	
	private boolean runningSimulation;
	
	public World(PrisonersNoRecognition owner_) { 
		owner = owner_;
		
		tournamentPairs = null;
		runningSimulation = true;
		
		//reset overall stats:
		totalGenerations = 0;
		minGenerations = 999999999;
		maxGenerations = 0;
		
		minUniqueStrategies = 999999999;
		maxUniqueStrategies = 0;
		totalUniqueStrategies = 0;
		
		minUniqueWinningStrategies = 999999999;
		maxUniqueWinningStrategies = 0;
		totalUniqueWinningStrategies = 0;	
		
		totalWinningFitness = 0;
		minWinningFitness = 999999999;
		maxWinningFitness = 0;
		
		totalAverageFitness = 0;
		minAverageFitness = 999999999;
		maxAverageFitness = 0;
			
		this.initEvolution();
		
	}
	
	public void initEvolution() {
		agentList = new ArrayList <Agent>();
		for (int i=0; i<NUM_AGENTS; i++) {
			agentList.add(new Agent(i, this));
		}	
		ReportController.getSingletonObject().resetEvolBased();
		generation = 0;
		initGeneration();
		
	}
	
	public void endEvolution() {
		evolution++;
		//-------- end-of-evol stats:
		if (OUTPUT_GENERATION_COOPERATIONS) {
			owner.getOutputPanel().displayTextInLargeOutput("");
			owner.getOutputPanel().displayTextInLargeOutput("<<<<<<<<<<<<<<<< END OF EVOL; generations: "+generation + " >>>>>>>>>>>>>>>");
		}
			
		totalGenerations += generation;
		if (generation < minGenerations) {
			minGenerations = generation;
		}
		if (generation > maxGenerations) {
			maxGenerations = generation;
		}
		
		
		double totalFitn = 0;
		//--- collate all strategies from sorted agent list to find unique, ordered by best:
		ArrayList<Agent> uniqueStrategyAgents = new ArrayList<Agent>();
		endOfEvolutionStrategies = new ArrayList <String>();
		for (int i=0; i<NUM_AGENTS; i++) {
			totalFitn += this.agentList.get(i).getFitness();
			String strategiesStr = this.agentList.get(i).translateDna(false);//this.agentList.get(i).getStrategiesRecord();
			
			if (!this.endOfEvolutionStrategies.contains(strategiesStr)) {
				this.endOfEvolutionStrategies.add(strategiesStr);
				uniqueStrategyAgents.add(this.agentList.get(i));
			}
		}
		//display strategies:
		if (OUTPUT_GENERATION_COOPERATIONS) {
			owner.getOutputPanel().displayTextInLargeOutput("UNIQUE STRATEGIES:");
		}
		int uniqueWinningStrategies = 0;
		int uniqueStrategies = uniqueStrategyAgents.size();
		for (int i=0; i<uniqueStrategyAgents.size(); i++) {
			if (OUTPUT_GENERATION_COOPERATIONS) {
				owner.getOutputPanel().displayTextInLargeOutput(uniqueStrategyAgents.get(i).translateDna(true)); //uniqueStrategyAgents.get(i).getStrategiesSummary()
			}
			//if the agent is amongst ones with best fitness, count its strategy into winning:
			if (uniqueStrategyAgents.get(i).getFitness() == this.agentList.get(0).getFitness()) {
				uniqueWinningStrategies++;	
			}
		}
		//add up strategies numbers:
		totalUniqueStrategies += uniqueStrategies;
		if (uniqueStrategies < minUniqueStrategies) {
			minUniqueStrategies = uniqueStrategies;
		}
		if (uniqueStrategies > maxUniqueStrategies) {
			maxUniqueStrategies = uniqueStrategies;
		}
		
		double percentageWinningStrategies = (double)uniqueWinningStrategies / (double) uniqueStrategies;
		
		totalUniqueWinningStrategies += percentageWinningStrategies;
		if (percentageWinningStrategies < minUniqueWinningStrategies) {
			minUniqueWinningStrategies = percentageWinningStrategies;
		}
		if (percentageWinningStrategies > maxUniqueWinningStrategies) {
			maxUniqueWinningStrategies = percentageWinningStrategies;
		}
		
		//------ fitness stats
		ArrayList <Agent> sortedList = new ArrayList<Agent>();
		sortedList.addAll(this.agentList);
		Collections.sort(sortedList);
		
		double winningFit = sortedList.get(0).getFitness();
		totalWinningFitness += winningFit;
		if (winningFit < minWinningFitness) {
			minWinningFitness = winningFit;
		}
		if (winningFit > maxWinningFitness) {
			maxWinningFitness = winningFit;
		}
		
		double averageFitn = (double)totalFitn / (double) NUM_AGENTS;
		totalAverageFitness += averageFitn;
		if (averageFitn < minAverageFitness) {
			minAverageFitness = averageFitn;
		}
		if (averageFitn > maxAverageFitness) {
			maxAverageFitness = averageFitn;
		}
		
		if (OUTPUT_GENERATION_COOPERATIONS) {
			owner.getOutputPanel().displayTextInLargeOutput("FITNESS: winning: "+winningFit+" average: "+averageFitn+" worst:"+this.agentList.get(NUM_AGENTS-1).getFitness());
			owner.getOutputPanel().displayTextInLargeOutput(""+totalFitn);
		
			owner.getOutputPanel().displayTextInLargeOutput("");
		}
		//---------
		if (evolution < NUM_EVOLUTIONS) {
			this.initEvolution();
		}
	}
	
	public void initGeneration() {		
		//reset counters:
		cooperationsCount = 0;
		actionsCount = 0;
		mutCooperationsCount = 0;
		tournamentNo = 0;
		
		String smallOutput = "EVOLUTION: " + (evolution+1) + " GENERATION: "+generation;
		smallOutput += "\nVALID EVOLUTIONS (ended by reaching the target cooperations): " + ReportController.getSingletonObject().numValidEvolutions;
		owner.getOutputPanel().displayTextInSmallOutput(smallOutput);
		if (OUTPUT_GENERATION_COOPERATIONS || OUTPUT_GENERATION_OTHER || OUTPUT_GENERATION_STRATEGIES || OUTPUT_TOURNAMENT) {
			owner.getOutputPanel().displayTextInLargeOutput("========================== GENERATION" + generation+ "===================");
			for (int i=0; i<NUM_AGENTS; i++)
			{
				//owner.getOutputPanel().displayTextInLargeOutput(this.agentList.get(i).translateDna(true));
			} 
		}
		
		//reset agents:
		for (int i=0; i< NUM_AGENTS; i++) {
			this.agentList.get(i).initState();
		}
		
	}
	
	/*
	 * update()
	 * - called in each frame of the progam
	 */
	
	public void update() {
		if (evolution < NUM_EVOLUTIONS) {
			if (tournamentNo < TOURNAMENTS_PER_GENERATION) {
				if (OUTPUT_TOURNAMENT) {
					owner.getOutputPanel().displayTextInLargeOutput("------------------ TOURNAMENT" + tournamentNo+ "------------------");
				}
				
				//-------------------- randomly generate pairs of agents that will play:
				tournamentPairs = generatePairs(0,agentList.size());					
				//-------------------- make them play Nth  tournament in M rounds
				for (int i=0; i<tournamentPairs.length; i+=2) {		
					Agent player1 = agentList.get((Integer)tournamentPairs[i]);
					Agent player2 = agentList.get((Integer)tournamentPairs[i+1]);
					player1.initGameState();
					player2.initGameState();
					int rnds = Maths.getSingletonObject().getRandomInteger(150,160);
					if (OUTPUT_TOURNAMENT) {
						owner.getOutputPanel().displayTextInLargeOutput("------------------------------------------");
					}
					for (int tournamentRound=0; tournamentRound<rnds; tournamentRound++) {
						String outputStr = "";
						if (OUTPUT_TOURNAMENT) {
							outputStr += "RND "+tournamentRound+ ", Player "+player1.getId()+ "(P"+player1.getId()+"="+player1.getPayoff()+")" + " vs "+player2.getId() + "(P"+player2.getId()+"="+player2.getPayoff()+")"; 
						}
						Agent.Action pl1action = player1.doAction();
						Agent.Action pl2action = player2.doAction();
						player1.registerActions(pl2action, pl1action);
						player2.registerActions(pl1action, pl2action);
						player1.addPlayedWith(player2.getId());// getAlreadyPlayedWithList().add();
						player2.addPlayedWith(player1.getId());
							
						if (pl1action == Agent.Action.COOPERATE) {
							cooperationsCount++;
						}
						if (pl2action == Agent.Action.COOPERATE) {
							cooperationsCount++;
						}		
						if (pl2action == Agent.Action.COOPERATE && pl1action == Agent.Action.COOPERATE) {
							mutCooperationsCount+=2;
						}
						actionsCount += 2;
						//---------compute payoffs and give them to players:
						byte[] payoffs = this.calculatePayoffs(pl1action, pl2action);
						player1.setPayoff(player1.getPayoff() + payoffs[0]);
						player2.setPayoff(player2.getPayoff() + payoffs[1]);
						//---------display output:
						if (OUTPUT_TOURNAMENT) {				
							outputStr += ":: A"+player1.getId()+"="+translateAction(pl1action) + "   A"+player2.getId()+"="+translateAction(pl2action) + " >> PAYOFFS: P"+player1.getId()+"="+player1.getPayoff() + "   P"+player2.getId()+"="+ player2.getPayoff();
							if (tournamentRound > 1) {
								if ((pl1action==Agent.Action.COOPERATE && player1.getInputs()[2] == 1) || (pl1action==Agent.Action.DEFECT && player1.getInputs()[2] == 0)) {
									//outputStr += " : PL1 TFT";
								} else {
									player1.setPlayingTFT(false);
								}
								if ((pl2action==Agent.Action.COOPERATE && player2.getInputs()[2] == 1) || (pl2action==Agent.Action.DEFECT && player2.getInputs()[2] == 0)) {
									//outputStr += " : PL2 TFT";
									player2.setPlayingTFT(false);
								}
							}
							owner.getOutputPanel().displayTextInLargeOutput(outputStr);
							/*owner.getOutputPanel().displayTextInLargeOutput(player1.translateDna(true));
							owner.getOutputPanel().displayTextInLargeOutput(player1.translateInputs());
							owner.getOutputPanel().displayTextInLargeOutput(player2.translateDna(true));
							owner.getOutputPanel().displayTextInLargeOutput(player2.translateInputs());  */	
						}
					}
					// end of game
					if (OUTPUT_TOURNAMENT) {
						owner.getOutputPanel().displayTextInLargeOutput("PL1:"  + player1.isPlayingTFT() + "PL2:" + player2.isPlayingTFT());
					}
					
				}
				if (OUTPUT_TOURNAMENT) {
					owner.getOutputPanel().displayTextInLargeOutput("");
				}
				// end of tournament
				tournamentNo++;
			} else {
				//====================================== all tournaments complete, END OF GENERATION ========================
				//percentage of cooperations out of total games played each generation
				double cooperationsPercent = (double)cooperationsCount / (double)actionsCount;// (double)cooperationsCount / (double)((tournamentPairs.length)*TOURNAMENTS_PER_GENERATION*TOURNAMENT_ROUNDS);
				double mutCooperationsPercent = (double)mutCooperationsCount / (double)actionsCount; //(double)((tournamentPairs.length)*TOURNAMENTS_PER_GENERATION*TOURNAMENT_ROUNDS);
				boolean endEvolution = (cooperationsPercent >= TARGET_COOPERATION_PERCENT) || (generation >= MAX_GENERATIONS);
				//-------------------------------------------- end-of-generation statistics
				double allFitness = 0;
				for (int i=0; i<NUM_AGENTS; i++) {
					Agent agent = this.agentList.get(i);
					allFitness += agent.getFitness();
					ReportController.getSingletonObject().registerStrategies(0, generation, i, NUM_AGENTS, agent, true, endEvolution);
				}
				ArrayList <Agent> sortedList = new ArrayList<Agent>();
				sortedList.addAll(this.agentList);
				Collections.sort(sortedList);
				// register the fitness:
				double averageFitn = (double)allFitness  / (double) NUM_AGENTS;
				ReportController.getSingletonObject().registerFitness(generation, sortedList.get(0).getFitness(), averageFitn, sortedList.get(NUM_AGENTS-1).getFitness());
				// register cooperation percentage:
				
				ReportController.getSingletonObject().registerCooperations(generation, cooperationsPercent, evolution);
				ReportController.getSingletonObject().registerMutualCooperations(generation, mutCooperationsPercent, evolution);
				//register best weights:
				ReportController.getSingletonObject().registerWeights(generation, sortedList.get(0).getGenotype());
				
				if (OUTPUT_GENERATION_OTHER) {
					owner.getOutputPanel().displayTextInLargeOutput(" BEST FITNESS: ("+sortedList.get(0).getId()+")"+ sortedList.get(0).getFitness());
					owner.getOutputPanel().displayTextInLargeOutput(" AVERAGE FITNESS: "+averageFitn);
					owner.getOutputPanel().displayTextInLargeOutput(" WINNER WEIGHTS: "+ sortedList.get(0).translateDna(true));
				}
				if (OUTPUT_GENERATION_COOPERATIONS) {
					owner.getOutputPanel().displayTextInLargeOutput(" COOPERATIONS (%): "+ cooperationsPercent);
					owner.getOutputPanel().displayTextInLargeOutput(" MUT COOPERATIONS (%): "+ mutCooperationsPercent);
				}
				
				//owner.getOutputPanel().displayTextInLargeOutput("TOTAL : "+ReportController.getSingletonObject().endOfGenCooperations);
				//owner.getOutputPanel().displayTextInLargeOutput("MUT : "+ReportController.getSingletonObject().endOfGenMutCooperations);
				
				if (OUTPUT_GENERATION_STRATEGIES) {
					owner.getOutputPanel().displayTextInLargeOutput(" BEST STRATEGIES: ");
					for (int i=0; i<5; i++) {
						if (i<NUM_AGENTS) {
							owner.getOutputPanel().displayTextInLargeOutput(sortedList.get(i).translateDna(true));
						}
					}
				} 
					
				//-------------------- end condition of evolution:
				if (endEvolution) {
					this.endEvolution();
				} else {
					//------------------------------------------------ create new generation:
					
					//--- remove last N agents so that NUM_REPLICATED_AGENTS remain:
					/*for (int i = agentList.size()-1; i> NUM_REPLICATED_AGENTS-1; i--) {
						agentList.remove(i);
					}
					//-- save the best agents into another list, null the agentList:
					ArrayList <Agent> parentList = agentList;
					agentList = new ArrayList<Agent>();
					//---- create new agents:
					for (int i=0; i<NUM_AGENTS; i++) {
						//-- create new parent pair:
						int parent1Id = Maths.getSingletonObject().getRandomInteger(0, parentList.size() - 1);
						int parent2Id = Maths.getSingletonObject().getRandomInteger(0, parentList.size() - 1);
						Agent parent1 =  parentList.get(parent1Id);
						Agent parent2 =  parentList.get(parent2Id);
						//-- create new agent based on genotype of the parents:
						Agent newAgent = new Agent(i,this,parent1 , parent2);
						agentList.add(newAgent);
					} 
					int a = 0; */
					for (int i = 0; i<NUM_REPLICATED_AGENTS; i++) {	
			    		//======  USING GEOGRAPHY: second parent from a deme where genot1 is in the middle: 
			    		int agentNo1 = Maths.getSingletonObject().getRandomInteger(0, NUM_AGENTS-1);//
			    		int agentNo2 = (int)(agentNo1 - (double)DEME_SIZE/2 + Maths.getSingletonObject().getRandomInteger(0,DEME_SIZE));
			    		//wrap-around:
			    		if (agentNo2 >= NUM_AGENTS) {
			    			agentNo2 = agentNo2 - NUM_AGENTS;
			    		} else if (agentNo2 < 0) {
			    			agentNo2 = NUM_AGENTS + agentNo2;
			    		} 
			    		Agent agent1 = agentList.get(agentNo1);
			    		Agent agent2 = agentList.get(agentNo2);
			    		
			    		Agent winner, looser;	    		
			    		if (agent1.getFitness() > agent2.getFitness()) {
			    			winner = agent1;
			    			looser = agent2;
			    		} else {
			    			looser = agent1;
			    			winner = agent2;
			    		}
			    		
			    		//compare the genotypes to determine which one will be passing genes, if both the same fitness keep them both:
			    		if (agent1.getFitness() != agent2.getFitness()) {		    		
				    		//change the genotype of looser according to winner:
				    		looser.changeGenotype(winner);
			    		} else {
			    			//make sure looser at least mutates:
			    			looser.changeGenotype(looser);
			    			//winner.changeGenotype(winner);
			    		}
		    		} 
					generation++;
					initGeneration();
				}
			}
		} else if (runningSimulation){
			this.endSimulation();
		}
		
	}
	
	
	public void endSimulation() {
		owner.getOutputPanel().displayTextInLargeOutput("");
		this.owner.getOutputPanel().displayTextInLargeOutput(">>>>>>>>>>>>>>>>>>>>>>>>>> END OF SIMULATION <<<<<<<<<<<<<<<<<<<<");
		double averGen = (double) totalGenerations / (double) NUM_EVOLUTIONS;
		this.owner.getOutputPanel().displayTextInLargeOutput("Generations: average: "+averGen + " max: "+maxGenerations + " min: "+minGenerations);
		
		double averStrat = (double) totalUniqueStrategies / (double) NUM_EVOLUTIONS;
		this.owner.getOutputPanel().displayTextInLargeOutput("Unique strategies: average: "+averStrat + " max: "+maxUniqueStrategies + " min: "+minUniqueStrategies);
		
		int averWinStrat = (int)Math.round( ((double) totalUniqueWinningStrategies / (double) NUM_EVOLUTIONS)*100);
		int maxWinStrat = (int)Math.round(maxUniqueWinningStrategies*100);
		int minWinStrat = (int)Math.round(minUniqueWinningStrategies*100);
		this.owner.getOutputPanel().displayTextInLargeOutput("Unique winning strategies: average: "+averWinStrat + "% max: "+ maxWinStrat + "% min: "+minWinStrat+"%");
		
		double averWinningFit = (double) totalWinningFitness / (double) NUM_EVOLUTIONS;
		this.owner.getOutputPanel().displayTextInLargeOutput("Winning fitness: average: "+averWinningFit + " max: "+maxWinningFitness + " min: "+minWinningFitness);
		
		double averAverFit = (double) totalAverageFitness / (double) NUM_EVOLUTIONS;
		this.owner.getOutputPanel().displayTextInLargeOutput("Average fitness: average: "+averAverFit + " max: "+maxAverageFitness + " min: "+minAverageFitness);
		
		this.owner.getOutputPanel().displayTextInLargeOutput("Average cooperations between gen. 100 & 400:" + ReportController.getSingletonObject().getEndOfGenerationCooperationsAverage());
		this.owner.getOutputPanel().displayTextInLargeOutput("Average mutual cooperations between gen. 100 & 400:" + ReportController.getSingletonObject().getEndOfGenerationMutCooperationsAverage());
		this.owner.getOutputPanel().displayTextInLargeOutput("Average max cooperations between gen. 100 & 400:" + ReportController.getSingletonObject().getEndOfGenerationCooperationsMaxAverage());
		this.owner.getOutputPanel().displayTextInLargeOutput("Average winning fitness between gen. 100 & 400:" + ReportController.getSingletonObject().getEndOfGenerationBestFitnessAverage());
		this.owner.getOutputPanel().displayTextInLargeOutput("Average average fitness between gen. 100 & 400:" + ReportController.getSingletonObject().getEndOfGenerationAverageFitnessAverage());
		this.owner.getOutputPanel().displayTextInLargeOutput("Number valid evolutions: "+ ReportController.getSingletonObject().getNumValidEvolutions());
		
		this.runningSimulation = false;
	}
	
	
	
	
	
	//=====================================================================
	
	private byte[] calculatePayoffs(Agent.Action pl1action, Agent.Action pl2action) {
		byte[] payoffs = new byte[2];
		if (pl1action == Agent.Action.COOPERATE) {
			if (pl2action == Agent.Action.COOPERATE) {
				//both cooperate, both receive 3:
				payoffs[0] = 3;
				payoffs[1] = 3;
			} else {
				//pl1 cooperates, pl2 defects -> pl2 receives 5, other 0:
				payoffs[0] = 0;
				payoffs[1] = 5;
			}
		} else {
			if (pl2action == Agent.Action.COOPERATE) {
				//pl2 cooperates, pl1 defects -> pl1 receives 5, other 0:
				payoffs[0] = 5;
				payoffs[1] = 0;
			} else {
				//both defect, both receive 1
				payoffs[0] = 1;
				payoffs[1] = 1;
			}
		}
		return payoffs;	
	}
	private Object[] generatePairs(int startNum, int endNum) {
		ArrayList<Integer> pairs = new ArrayList<Integer>();
		while (pairs.size() < NUM_AGENTS) {
			Integer nextInt = Integer.valueOf(Maths.getSingletonObject().getRandomInteger(startNum, endNum-1) + startNum);
			if (!pairs.contains(nextInt)) {
				boolean canAdd = true;
				if (pairs.size() %2 == 1 ) {
					Integer a = pairs.get(pairs.size()-1);
					Agent player1 = agentList.get(a);
					//make sure this pair did not play in this generation:
					if (player1.alreadyPlayedWith(nextInt)) {
						canAdd = false;
						counter++;
						if (counter == 10) {
							pairs = new ArrayList<Integer>(); //try other combinations...
							canAdd = true;
						}
					}
				}
				if (canAdd)  {
					counter = 0;
					pairs.add(nextInt);
				}
			}
		}
		//this.owner.getOutputPanel().displayTextInLargeOutput("=============================== PAIRS ");
		return (Object[]) pairs.toArray();
	}
	
	public String translateAction(Agent.Action action) {
		if (action == Agent.Action.COOPERATE) {
			return "C";
		} else {
			return "D";
		}
	}
	
	//=================================================================================== GETTERS / SETTERS
	
	public int getNumAgents() {
		return agentList.size();
	}
	
	public Agent getAgentsWithId(int id_) {
		return agentList.get(id_);
	}
	
	public List <Agent> getAgentList() {
		return agentList;
	}


	
	

	


}
