/*
 * Downloaded from lenkaspace.net
 * 
 * Paper about this project: http://lenkaspace.net/papers/iteratedPrisonersDilemmaNeuralNetworks.php
 * Email: contact@lenkaspace.net
 */

package net.lenkaspace.prisonersnorecognition.reports;

import java.awt.Dimension;
import java.util.ArrayList;
import java.util.List;

import net.lenkaspace.prisonersnorecognition.model.Agent;
import net.lenkaspace.prisonersnorecognition.model.World;

import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;


public class ReportController {
	
	
	//---------------- CONSTANTS
	public static final int REPORT_FITNESS = 0;
	public static final int REPORT_COOPERATIONS = 1;
	public static final int REPORT_WEIGHTS = 2;
	public static final int REPORT_COOPERATIVE_STRATEGIES= 3;
	public static final int REPORT_DEFECTIVE_STRATEGIES= 4;
	public static final int REPORT_END_COOP_STRATEGIES = 5;
	public static final int REPORT_END_DEF_STRATEGIES = 6;
	public static final int REPORT_MUT_COOPERATIONS = 7;
	public static final int REPORT_END_OF_GEN_COOPERATIONS = 8;
	public static final int REPORT_END_OF_GEN_MUT_COOPERATIONS = 9;
	public static final int REPORT_END_OF_GEN_COOP_STRATEGIES = 10;
	public static final int REPORT_END_OF_GEN_DEF_STRATEGIES = 11;
	
	//------------------------
	//private static final int NUM_STRATEGIES = 4;
	public static final int NUM_STRATEGIES = 16;
	
	//private static final String[] cooperativeStrategiesNames = {"C","(C) C", "(D) C", "(C,C) C", "(C,D) C", "(D,C) C", "(D,D) C" };
	//private static final String[] defectiveStrategiesNames = {"D","(C) D", "(D) D", "(C,C) D", "(C,D) D", "(D,C) D", "(D,D) D" };
	
	//private static final String[] cooperativeStrategiesNames = {"C DC","C DD", "C CC", "C CD" };
	private static final String[] cooperativeStrategiesNames = {"[C CCCC]", "[C CCDC]", "[C CCCD]", "[C CCDD]", "C CDCC", "C CDDC", "C CDCD", "C CDDD", "C DCCC", "C DCDC", "C DCCD", "C DCDD", "C DDCC", "C DDDC", "C DDCD", "C DDDD"};
	//private static final String[] defectiveStrategiesNames = {"D DC","D DD", "D CC", "D CD" };
	private static final String[] defectiveStrategiesNames = {"D CCCC", "D CCDC", "D CCCD", "D CCDD", "D CDCC", "D CDDC", "D CDCD", "D CDDD", "[D DCCC]", "[D DCDC]", "[D DCCD]", "[D DCDD]", "D DDCC", "D DDDC", "D DDCD", "-D DDDD-"};
	
	private int[] cooperativeStrategiesCounter = new int[NUM_STRATEGIES];
	private int[] defectiveStrategiesCounter = new int[NUM_STRATEGIES];
	
	//--------------------------
	private List<XYSeries> dataFitness;
	private List<XYSeries> dataCooperation;
	private List<XYSeries> dataWeights;
	private List<XYSeries> dataCooperationStrategies;
	private List<XYSeries> dataDefectionStrategies;
	private List<XYSeries> dataMutualCooperations;
	
	private List<XYSeries> dataEndCoopStrategies; //at the end of each evol, register how many agents had certain strategy
	private List<XYSeries> dataEndDefStrategies;
	/*private double[] endStrategiesCounterTotal = new double[NUM_STRATEGIES*2];
	private double[] endStrategiesCounterMin = new double[NUM_STRATEGIES*2];
	private double[] endStrategiesCounterMax = new double[NUM_STRATEGIES*2];
	private int [] endStrategiesCounterThisEvol = new int[NUM_STRATEGIES*2];*/
	private double[] endCoopStrategiesCounterTotal = new double[NUM_STRATEGIES];
	private double[] endCoopStrategiesCounterMin = new double[NUM_STRATEGIES];
	private double[] endCoopStrategiesCounterMax = new double[NUM_STRATEGIES];
	private int [] endCoopStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
	
	private double[] endDefStrategiesCounterTotal = new double[NUM_STRATEGIES];
	private double[] endDefStrategiesCounterMin = new double[NUM_STRATEGIES];
	private double[] endDefStrategiesCounterMax = new double[NUM_STRATEGIES];
	private int [] endDefStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
	
	private List<XYSeries> dataEndOfGenCoopStrategies;
	private double[] endOfGenCoopStrategiesCounterTotal = new double[NUM_STRATEGIES];
	private double[] endOfGenCoopStrategiesCounterMin = new double[NUM_STRATEGIES];
	private double[] endOfGenCoopStrategiesCounterMax = new double[NUM_STRATEGIES];
	private int [] endOfGenCoopStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
	
	private List<XYSeries> dataEndOfGenDefStrategies;
	private double[] endOfGenDefStrategiesCounterTotal = new double[NUM_STRATEGIES];
	private double[] endOfGenDefStrategiesCounterMin = new double[NUM_STRATEGIES];
	private double[] endOfGenDefStrategiesCounterMax = new double[NUM_STRATEGIES];
	private int [] endOfGenDefStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
	
	private double mutualCoopsCounterTotal;
	
	private double endOfGenCooperations;
	private double endOfGenCooperationsMax;
	private double endOfGenCooperationsMin;
	private double endOfGenMutCooperations;
	private double endOfGenMutCooperationsMax;
	private double endOfGenMutCooperationsMin;
	private double endOfGenCooperationsMaxAverage = 0;
	private double endOfGenCooperationsMaxMin = 0;
	private double endOfGenCooperationsMaxMax = 0;
	private double endOfGenCooperationsMaxThis = 0;
	private double endOfGenBestFitness = 0;
	private double endOfGenBestFitnessMax = 0;
	private double endOfGenBestFitnessMin = 0;
	private double endOfGenFitness =0;
	private double endOfGenFitnessMax =0;
	private double endOfGenFitnessMin =0;
	
	public int numValidEvolutions;
	
	private BaseReport baseReport;
	

	
	private static ReportController ref;
	
	 
	
	private ReportController() 	{
		reset();
		baseReport = new BaseReport(new Dimension(950,400));
		//create array of input combinations for using NN:
		
		
	}
	
	public static ReportController getSingletonObject()	  {
	  if (ref == null)
	        // it's ok, we can call this constructor
	     ref = new ReportController();		
	  return ref;
	}
	
	public Object clone() throws CloneNotSupportedException {
	    throw new CloneNotSupportedException(); 
	}
	
	
	public void reset() {
		endCoopStrategiesCounterTotal = new double[NUM_STRATEGIES];
		endCoopStrategiesCounterMin = new double[NUM_STRATEGIES];
		for (int i=0; i<endCoopStrategiesCounterMin.length; i++) {
			endCoopStrategiesCounterMin[i] = 99999999.0;
		}
		endCoopStrategiesCounterMax = new double[NUM_STRATEGIES];
		endCoopStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
		dataEndCoopStrategies = new ArrayList<XYSeries>();
		
		endDefStrategiesCounterTotal = new double[NUM_STRATEGIES];
		endDefStrategiesCounterMin = new double[NUM_STRATEGIES];
		for (int i=0; i<endCoopStrategiesCounterMin.length; i++) {
			endDefStrategiesCounterMin[i] = 99999999.0;
		}
		endDefStrategiesCounterMax = new double[NUM_STRATEGIES];
		endDefStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
		dataEndDefStrategies = new ArrayList<XYSeries>();
		
		
		endOfGenCoopStrategiesCounterTotal = new double[NUM_STRATEGIES];
		endOfGenCoopStrategiesCounterMin = new double[NUM_STRATEGIES];
		for (int i=0; i<endOfGenCoopStrategiesCounterMin.length; i++) {
			endOfGenCoopStrategiesCounterMin[i] = 99999999.0;
		}
		endOfGenCoopStrategiesCounterMax = new double[NUM_STRATEGIES];
		endOfGenCoopStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
		dataEndOfGenCoopStrategies = new ArrayList<XYSeries>();
		
		endOfGenDefStrategiesCounterTotal = new double[NUM_STRATEGIES];
		endOfGenDefStrategiesCounterMin = new double[NUM_STRATEGIES];
		for (int i=0; i<endOfGenDefStrategiesCounterMin.length; i++) {
			endOfGenDefStrategiesCounterMin[i] = 99999999.0;
		}
		endOfGenDefStrategiesCounterMax = new double[NUM_STRATEGIES];
		endOfGenDefStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
		dataEndOfGenDefStrategies = new ArrayList<XYSeries>();
		
		
		dataMutualCooperations = new ArrayList <XYSeries>();
		mutualCoopsCounterTotal = 0;
		
		endOfGenCooperations = 0;
		endOfGenCooperationsMax = 0;
		endOfGenCooperationsMin = 99999999;
		endOfGenMutCooperations = 0;
		endOfGenMutCooperationsMax = 0;
		endOfGenMutCooperationsMin = 99999999;
		endOfGenCooperationsMaxAverage = 0;
		endOfGenCooperationsMaxMax = 0;
		endOfGenCooperationsMaxMin = 99999999;
		endOfGenCooperationsMaxThis = 0;
		endOfGenBestFitness = 0;
		endOfGenBestFitnessMax = 0;
		endOfGenBestFitnessMin = 99999999;
		endOfGenFitness = 0;
		endOfGenFitnessMax =0;
		endOfGenFitnessMin =99999999;
		
		numValidEvolutions = 0;
		
	}
	
	
	public void resetEvolBased() {
		dataFitness = new ArrayList<XYSeries>();
		dataCooperation = new ArrayList<XYSeries>();
		dataWeights = new ArrayList<XYSeries>();
		dataCooperationStrategies = new ArrayList<XYSeries>();
		dataDefectionStrategies = new ArrayList<XYSeries>();
		
	}
	
	//=================================== showing reports:
	
	public void showReport(int reportType_) {
		
		switch (reportType_) {
			case REPORT_FITNESS:
			
				baseReport.display("Best fitness", "Generation", "Fitness", this.prepareXYSeriesForDisplay(this.dataFitness));
				break;
			case REPORT_COOPERATIONS:
				baseReport.display("Cooperations (% out of total games played per generation)", "Generation", "Cooperations", this.prepareXYSeriesForDisplay(this.dataCooperation));
				break;
			case REPORT_WEIGHTS:
				baseReport.display("Weights", "Generation", "Fitness", this.prepareXYSeriesForDisplay(this.dataWeights));
				break;
			case REPORT_COOPERATIVE_STRATEGIES:
				baseReport.display("Cooperative Strategies", "Generation", "Strategy counter", this.prepareXYSeriesForDisplay(this.dataCooperationStrategies));
				break;
			case REPORT_DEFECTIVE_STRATEGIES:
				baseReport.display("Defective Strategies", "Generation", "Strategy counter", this.prepareXYSeriesForDisplay(this.dataDefectionStrategies));
				break;
			case REPORT_END_COOP_STRATEGIES:
				//add the counter data into the XYSeries:
				this.dataEndCoopStrategies = new ArrayList<XYSeries>();
				XYSeries series = new XYSeries("Max");
				series.setDescription("Max");
				this.dataEndCoopStrategies.add(series);
	
				XYSeries series2 = new XYSeries("Average");
				series2.setDescription("Average");
				this.dataEndCoopStrategies.add(series2);
				
				XYSeries series3 = new XYSeries("Min");
				series3.setDescription("Min");
				this.dataEndCoopStrategies.add(series3);
				
				
				for (int i=0; i<endCoopStrategiesCounterTotal.length;i++) {
					this.dataEndCoopStrategies.get(0).add(i,endCoopStrategiesCounterMax[i]); //max
					this.dataEndCoopStrategies.get(1).add(i,(double)endCoopStrategiesCounterTotal[i] / (double)numValidEvolutions); //average
					this.dataEndCoopStrategies.get(2).add(i,endCoopStrategiesCounterMin[i]); //min
				}
				baseReport.display("End cooperation strategies", "[CCCC]   [CCDC]   [CCCD]   [CCDD]     CDCC      CDDC      CDCD     CDDD     DCCC      DCDC     DCCD      DCDD     DDCC     DDDC     DDCD     DDDD           ", "Strategy counter (%/agents)", this.prepareXYSeriesForDisplay(this.dataEndCoopStrategies));
				//baseReport.display("End strategies", "	C          (C)C       (D)C      (C,C)C      (C,D)C      (D,C)C      (D,D)C         D           (C)D        (D)D       (C,C)D      (C,D)D      (D,C)D      (D,D)D", "Strategy counter (%/agents)", this.prepareXYSeriesForDisplay(this.dataEndCoopStrategies));
				break;
			case REPORT_END_DEF_STRATEGIES:
				//add the counter data into the XYSeries:
				this.dataEndDefStrategies = new ArrayList<XYSeries>();
				XYSeries series4 = new XYSeries("Max");
				series4.setDescription("Max");
				this.dataEndDefStrategies.add(series4);
	
				XYSeries series5 = new XYSeries("Average");
				series5.setDescription("Average");
				this.dataEndDefStrategies.add(series5);
				
				XYSeries series6 = new XYSeries("Min");
				series6.setDescription("Min");
				this.dataEndDefStrategies.add(series6);
				
				
				for (int i=0; i<endDefStrategiesCounterTotal.length;i++) {
					this.dataEndDefStrategies.get(0).add(i,endDefStrategiesCounterMax[i]); //max
					this.dataEndDefStrategies.get(1).add(i,(double)endDefStrategiesCounterTotal[i] / (double)numValidEvolutions); //average
					this.dataEndDefStrategies.get(2).add(i,endDefStrategiesCounterMin[i]); //min
				}
				baseReport.display("End defection strategies", "CCCC     CCDC     CCCD       CCDD     CDCC     CDDC      CDCD      CDDD   [DCCC]     [DCDC]   [DCCD]   [DCDD]    DDCC     DDDC     DDCD    DDDD           ", "Strategy counter (%/agents)", this.prepareXYSeriesForDisplay(this.dataEndDefStrategies));
				
				break;
			case REPORT_MUT_COOPERATIONS:
				baseReport.display("Mutual Cooperations (average over evolution)", "Generation", "Mutual Cooperations", this.prepareXYSeriesForDisplay(this.dataMutualCooperations));
				break;
			/*case REPORT_END_OF_GEN_COOPERATIONS:
				double average = this.endOfGenCooperations / (100*World.NUM_EVOLUTIONS);
				break;
			case REPORT_END_OF_GEN_COOPERATIONS:
				double average = this.endOfGenCooperations / (100*World.NUM_EVOLUTIONS);
				break;*/
			case REPORT_END_OF_GEN_COOP_STRATEGIES:
				//add the counter data into the XYSeries:
				this.dataEndOfGenCoopStrategies = new ArrayList<XYSeries>();
				XYSeries series7 = new XYSeries("Max");
				series7.setDescription("Max");
				this.dataEndOfGenCoopStrategies.add(series7);
	
				XYSeries series8 = new XYSeries("Average");
				series8.setDescription("Average");
				this.dataEndOfGenCoopStrategies.add(series8);
				
				XYSeries series9 = new XYSeries("Min");
				series9.setDescription("Min");
				this.dataEndOfGenCoopStrategies.add(series9);
				
				XYSeries series14 = new XYSeries("-");
				series14.setDescription("-");
				this.dataEndOfGenCoopStrategies.add(series14);
				
				
				for (int i=0; i<endOfGenCoopStrategiesCounterTotal.length;i++) {
					this.dataEndOfGenCoopStrategies.get(0).add(i,endOfGenCoopStrategiesCounterMax[i]); //max
					this.dataEndOfGenCoopStrategies.get(1).add(i,(double)endOfGenCoopStrategiesCounterTotal[i] / (double)(World.NUM_EVOLUTIONS)); //average over N evolutions , 300 rounds each
					this.dataEndOfGenCoopStrategies.get(2).add(i,endOfGenCoopStrategiesCounterMin[i]); //min
					this.dataEndOfGenCoopStrategies.get(3).add(i,1); 
				}
				baseReport.display("Cooperation strategies between gen. 100-400", "[CCCC]   [CCDC]   [CCCD]   [CCDD]     CDCC      CDDC      CDCD     CDDD     DCCC      DCDC     DCCD      DCDD     DDCC     DDDC     DDCD     DDDD           ", "Strategy counter (%/agents)", this.prepareXYSeriesForDisplay(this.dataEndOfGenCoopStrategies));
				//baseReport.display("End strategies", "	C          (C)C       (D)C      (C,C)C      (C,D)C      (D,C)C      (D,D)C         D           (C)D        (D)D       (C,C)D      (C,D)D      (D,C)D      (D,D)D", "Strategy counter (%/agents)", this.prepareXYSeriesForDisplay(this.dataEndCoopStrategies));
				break;
			case REPORT_END_OF_GEN_DEF_STRATEGIES:
				//add the counter data into the XYSeries:
				this.dataEndOfGenDefStrategies = new ArrayList<XYSeries>();
				XYSeries series10 = new XYSeries("Max");
				series10.setDescription("Max");
				this.dataEndOfGenDefStrategies.add(series10);
	
				XYSeries series11 = new XYSeries("Average");
				series11.setDescription("Average");
				this.dataEndOfGenDefStrategies.add(series11);
				
				XYSeries series12 = new XYSeries("Min");
				series12.setDescription("Min");
				this.dataEndOfGenDefStrategies.add(series12);
				
				XYSeries series13 = new XYSeries("-");
				series13.setDescription("-");
				this.dataEndOfGenDefStrategies.add(series13);
				
				
				for (int i=0; i<endCoopStrategiesCounterTotal.length;i++) {
					this.dataEndOfGenDefStrategies.get(0).add(i,endOfGenDefStrategiesCounterMax[i]); //max
					this.dataEndOfGenDefStrategies.get(1).add(i,(double)endOfGenDefStrategiesCounterTotal[i] / (double)(World.NUM_EVOLUTIONS)); //average over N evolutions , 300 rounds each
					this.dataEndOfGenDefStrategies.get(2).add(i,endOfGenDefStrategiesCounterMin[i]); //min
					this.dataEndOfGenDefStrategies.get(3).add(i,1); 
				}
				baseReport.display("Defection strategies between gen. 100-400", "[CCCC]   [CCDC]   [CCCD]   [CCDD]     CDCC      CDDC      CDCD     CDDD     DCCC      DCDC     DCCD      DCDD     DDCC     DDDC     DDCD     DDDD           ", "Strategy counter (%/agents)", this.prepareXYSeriesForDisplay(this.dataEndOfGenDefStrategies));
				
				//baseReport.display("End strategies", "	C          (C)C       (D)C      (C,C)C      (C,D)C      (D,C)C      (D,D)C         D           (C)D        (D)D       (C,C)D      (C,D)D      (D,C)D      (D,D)D", "Strategy counter (%/agents)", this.prepareXYSeriesForDisplay(this.dataEndCoopStrategies));
				break;
			default:
				break;
		}
	
	}
	
	public String getEndOfGenerationCooperationsAverage() {
		double average = this.endOfGenCooperations / (double)(300*World.NUM_EVOLUTIONS); //was measured between gen. 100-400
		return "Max: " + this.endOfGenCooperationsMax + " Average: " + average + " Min: "+this.endOfGenCooperationsMin;
	}
	public String getEndOfGenerationMutCooperationsAverage() {
		double average = this.endOfGenMutCooperations / (double)(300*World.NUM_EVOLUTIONS);
		return "Max: " + this.endOfGenMutCooperationsMax + " Average: " + average + " Min: "+this.endOfGenMutCooperationsMin;
	}
	
	public String getEndOfGenerationCooperationsMaxAverage() {
		double average = this.endOfGenCooperationsMaxAverage / (double)(World.NUM_EVOLUTIONS); // max value was added at the end of each evol
		return "Max: " + this.endOfGenCooperationsMaxMax + " Average: " + average + " Min: "+this.endOfGenCooperationsMaxMin;
	}
	public String getEndOfGenerationBestFitnessAverage() {
		double average = this.endOfGenBestFitness / (double)(300*World.NUM_EVOLUTIONS); // max value was added at the end of each evol
		return "Max: " + this.endOfGenBestFitnessMax + " Average: " + average + " Min: "+this.endOfGenBestFitnessMin;
	}
	
	public String getEndOfGenerationAverageFitnessAverage() {
		double average = this.endOfGenFitness / (double)(300*World.NUM_EVOLUTIONS); // max value was added at the end of each evol
		return "Max: " + this.endOfGenFitnessMax + " Average: " + average + " Min: "+this.endOfGenFitnessMin;
	}
	

	
	
	
	
	
	private XYDataset prepareXYSeriesForDisplay(List<XYSeries> rawData) {
		
		XYSeriesCollection dataset = new XYSeriesCollection();
		dataset.addSeries(new XYSeries("Legend")); //needs a series at the beginning, otherwise won't display...
		
		//go through the list of rawData (e.g. multiple variables the report should show:
		for (int var=0; var < rawData.size(); var++) {
			XYSeries rawSeries = rawData.get(var);
			XYSeries displaySeries = new XYSeries(rawSeries.getDescription());
			int dataSize = rawSeries.getItemCount();
			int desiredDataSize = 60;
			if (dataSize <= desiredDataSize) {
				try {
					displaySeries = (XYSeries) rawSeries.clone();
				} catch (CloneNotSupportedException e) {
					e.printStackTrace();
				}
			} else {
				for (int i = 0; i<desiredDataSize-1; i++) {
					//make sure it marks every Nth datapoint of original data, so that total num of X-axis points is dataSize=60
					int k = (int)Math.ceil((double)dataSize/(double)(desiredDataSize-1))*(i);
					if (k<dataSize) {
						displaySeries.add(k,rawSeries.getY(k));
					}			
				}
				//----set last value of graph's data set to be the last value of the original data:
				displaySeries.add(dataSize, rawSeries.getY(dataSize-1));
				//-----
			} 
			
			dataset.addSeries(displaySeries);  
		
		}
			        
	    return dataset;
	}
	

	//========================== data setting methods:
	public void registerMutualCooperations(int gen_, double mutCoops_, int evol_) {
		mutualCoopsCounterTotal += mutCoops_;
		
		if (evol_ == World.NUM_EVOLUTIONS-1) {
			//end of all evolutions, create data set and add data
			this.dataMutualCooperations = new ArrayList<XYSeries>();
			XYSeries series4 = new XYSeries("Average");
			series4.setDescription("Average");
			this.dataMutualCooperations.add(series4);
			
			series4.add(gen_,(double)mutualCoopsCounterTotal / (double)World.NUM_EVOLUTIONS); //average
			
			XYSeries series2 = new XYSeries("border");
			series2.setDescription("border");
			this.dataMutualCooperations.add(series2);
			series2.add(gen_, 1);
		}
		
		if (gen_ > 100 && gen_ <=400) {
			this.endOfGenMutCooperations += mutCoops_;
			if (this.endOfGenMutCooperationsMax < mutCoops_) {
				this.endOfGenMutCooperationsMax = mutCoops_;
			}
			if (this.endOfGenMutCooperationsMin > mutCoops_) {
				this.endOfGenMutCooperationsMin = mutCoops_;
			}
		}
		
		
	}
	
	public void registerCooperations(int gen_, double cooperationsPercent, int evolution) {
		if (dataCooperation.size() == 0) {
			XYSeries series = new XYSeries("Cooperation");
			series.setDescription("Cooperation");
			dataCooperation.add(series);
			XYSeries series2 = new XYSeries("border");
			series2.setDescription("border");
			dataCooperation.add(series2);
		}
		dataCooperation.get(0).add(gen_,cooperationsPercent);
		dataCooperation.get(1).add(gen_,1);
		
		if (gen_ > 100 && gen_ <=400) {
			this.endOfGenCooperations += cooperationsPercent;
			if (this.endOfGenCooperationsMax < cooperationsPercent) {
				this.endOfGenCooperationsMax = cooperationsPercent;
			}
			if (this.endOfGenCooperationsMin > cooperationsPercent) {
				this.endOfGenCooperationsMin = cooperationsPercent;
			}
			
			if (cooperationsPercent > this.endOfGenCooperationsMaxThis) {
				this.endOfGenCooperationsMaxThis = cooperationsPercent;
			}
			//if this was last record of this evol, add max to the average max and reset it to 0:
			if (gen_ == 400) {
				this.endOfGenCooperationsMaxAverage += this.endOfGenCooperationsMaxThis;
				if (this.endOfGenCooperationsMaxMax < this.endOfGenCooperationsMaxThis) {
					this.endOfGenCooperationsMaxMax = this.endOfGenCooperationsMaxThis;
				}
				if (this.endOfGenCooperationsMaxMin > this.endOfGenCooperationsMaxThis) {
					this.endOfGenCooperationsMaxMin = this.endOfGenCooperationsMaxThis;
				}
				
				this.endOfGenCooperationsMaxThis = 0;
			}
		}
	}
	
	
	public void registerFitness(int gen_, double bestFit_, double averageFit_, double worstFitn_) {
		if (dataFitness.size() == 0) {
			XYSeries series = new XYSeries("Best fitness");
			series.setDescription("Best fitness");
			dataFitness.add(series);
			
			XYSeries series2 = new XYSeries("Average fitness");
			series2.setDescription("Average fitness");
			dataFitness.add(series2);
			
			XYSeries series3 = new XYSeries("Worst fitness");
			series3.setDescription("Worst fitness");
			dataFitness.add(series3);
		}
		dataFitness.get(0).add(gen_, bestFit_);
		dataFitness.get(1).add(gen_, averageFit_);
		dataFitness.get(2).add(gen_, worstFitn_);
		
		if (gen_ > 100 && gen_ <= 400) {
			this.endOfGenBestFitness += bestFit_;
			if (this.endOfGenBestFitnessMax < bestFit_) {
				this.endOfGenBestFitnessMax = bestFit_;
			}
			if (this.endOfGenBestFitnessMin > bestFit_) {
				this.endOfGenBestFitnessMin = bestFit_;
			}
			this.endOfGenFitness += averageFit_;
			if (this.endOfGenFitnessMax < bestFit_) {
				this.endOfGenFitnessMax = bestFit_;
			}
			if (this.endOfGenFitnessMin > worstFitn_) {
				this.endOfGenFitnessMin = worstFitn_;
			}
		}
		
	}
	
	
	
	public void registerWeights(int gen_, double[] weights_) {
		/*if (dataWeights.size() == 0) {
			for (int i=0; i<weights_.length; i++) {
				XYSeries series = new XYSeries(Agent.phenotypeNames[i]);
				series.setDescription(Agent.phenotypeNames[i]);
				dataWeights.add(series);
			}
		}
		
		for (int i=0; i<weights_.length; i++) {
			dataWeights.get(i).add(gen_,weights_[i]);
		}*/
		
		
		/*Double[] weights = new Double[weights_.size()];
		for (int i=0; i< weights_.size(); i++) {
			weights[i] = (double)(Math.round(weights_.get(i)*100));
		}
		dataWeights.add(weights); (*/
	}
	
	/*
	 * end strategies: 
	 * 	0	1		2		3		4		5		6		7	8		9		10		11		12		13
	 * 	C	(C)C	(D)C	(C,C)C	(C,D)C	(D,C)C	(D,D)C	D	(C)D	(D)D	(C,C)D	(C,D)D	(D,C)D	(D,D)D
	 */
	
	public void registerStrategies (int tournNo, int gen_, int agentNo,int maxAgents,  Agent agent, boolean lastTestInTournament, boolean endEvolution) {
		if (this.dataCooperationStrategies.size() == 0) {
			for (int i=0; i<NUM_STRATEGIES; i++) {
				XYSeries series = new XYSeries(ReportController.cooperativeStrategiesNames[i]);
				series.setDescription(ReportController.cooperativeStrategiesNames[i]);
				this.dataCooperationStrategies.add(series);
			}
		}
		if (this.dataDefectionStrategies.size() == 0) {
			for (int i=0; i<NUM_STRATEGIES; i++) {
				XYSeries series = new XYSeries(ReportController.defectiveStrategiesNames[i]);
				series.setDescription(ReportController.defectiveStrategiesNames[i]);
				this.dataDefectionStrategies.add(series);
			}
		}
		
		double [] genotype = null;
		if (Agent.USING_NN) {
			genotype = agent.getGenotypeAsStratFromNN();
			
			
		} else {
			genotype = agent.getGenotype();
			
		}
		
		
		//evaluate actions / genotype:
		for (int i=0; i<NUM_STRATEGIES; i++) {
			if (genotype[i] == 1) {
				cooperativeStrategiesCounter[i] ++; 
				
				if (gen_   > 100 && gen_ <= 400) {
					endOfGenCoopStrategiesCounterThisEvol[i] ++;
				}
				if (endEvolution && gen_ < World.MAX_GENERATIONS) { 
					//last test in this evolution, register strategy to end strategies as well:
					endCoopStrategiesCounterThisEvol[i]++;
					
				}
			} else {
				defectiveStrategiesCounter[i] ++; 
				if (gen_   > 100 && gen_ <= 400) {
					endOfGenDefStrategiesCounterThisEvol[i] ++;
				}
				if (endEvolution && gen_ < World.MAX_GENERATIONS) { 
					//last test in this evolution, register strategy to end strategies as well:
					endDefStrategiesCounterThisEvol[i]++;
					
				}
			}
		}
		
		if (agentNo == maxAgents-1) {
			//last agent of the generation:
			for (int i=0; i<NUM_STRATEGIES; i++) {
				//if (i<4) {
					dataCooperationStrategies.get(i).add(gen_,cooperativeStrategiesCounter[i]);	
				//}
				cooperativeStrategiesCounter[i] = 0;
				//if (i>=8 && i < 12) {
					dataDefectionStrategies.get(i).add(gen_,defectiveStrategiesCounter[i]);
				//}
				defectiveStrategiesCounter[i] = 0;
			}
			
			if (gen_ == 400) {
				//last agent and last test in this evolution (if max. gen = 400), handle end-of-evolution strategies:
				for (int i=0; i<endOfGenCoopStrategiesCounterThisEvol.length; i++) {
					double percentageAgentDidStrategy = (double)endOfGenCoopStrategiesCounterThisEvol[i] / (double)(300* maxAgents);
					endOfGenCoopStrategiesCounterTotal[i] += percentageAgentDidStrategy;
					if (percentageAgentDidStrategy < endOfGenCoopStrategiesCounterMin[i]) {
						endOfGenCoopStrategiesCounterMin[i] = percentageAgentDidStrategy;
					}
					if (percentageAgentDidStrategy > endOfGenCoopStrategiesCounterMax[i]) {
						endOfGenCoopStrategiesCounterMax[i] = percentageAgentDidStrategy;
					} 
				}
				//reset end strategies counter:
				endOfGenCoopStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
				
				
				for (int i=0; i<endOfGenDefStrategiesCounterThisEvol.length; i++) {
					double percentageAgentDidStrategy = (double)endOfGenDefStrategiesCounterThisEvol[i] / (double) (300*maxAgents);
					endOfGenDefStrategiesCounterTotal[i] += percentageAgentDidStrategy;
					if (percentageAgentDidStrategy < endOfGenDefStrategiesCounterMin[i]) {
						endOfGenDefStrategiesCounterMin[i] = percentageAgentDidStrategy;
					}
					if (percentageAgentDidStrategy > endOfGenDefStrategiesCounterMax[i]) {
						endOfGenDefStrategiesCounterMax[i] = percentageAgentDidStrategy;
					} 
				}
				//reset end strategies counter:
				endOfGenDefStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
			}
			
			if (endEvolution) { 
				if (gen_ < World.MAX_GENERATIONS) { //only register strategies if evol didn't stop because of max generations reached
					numValidEvolutions ++;
					//last agent and last test in this evolution, handle end-of-evolution strategies:
					for (int i=0; i<endCoopStrategiesCounterThisEvol.length; i++) {
						double percentageAgentDidStrategy = (double)endCoopStrategiesCounterThisEvol[i] / (double)maxAgents;
						endCoopStrategiesCounterTotal[i] += percentageAgentDidStrategy;
						if (percentageAgentDidStrategy < endCoopStrategiesCounterMin[i]) {
							endCoopStrategiesCounterMin[i] = percentageAgentDidStrategy;
						}
						if (percentageAgentDidStrategy > endCoopStrategiesCounterMax[i]) {
							endCoopStrategiesCounterMax[i] = percentageAgentDidStrategy;
						} 
					}
				}
				//reset end strategies counter:
				endCoopStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
				
				if (gen_ < World.MAX_GENERATIONS) { //only register strategies if evol didn't stop because of max generations reached
					for (int i=0; i<endDefStrategiesCounterThisEvol.length; i++) {
						double percentageAgentDidStrategy = (double)endDefStrategiesCounterThisEvol[i] / (double)maxAgents;
						endDefStrategiesCounterTotal[i] += percentageAgentDidStrategy;
						if (percentageAgentDidStrategy < endDefStrategiesCounterMin[i]) {
							endDefStrategiesCounterMin[i] = percentageAgentDidStrategy;
						}
						if (percentageAgentDidStrategy > endDefStrategiesCounterMax[i]) {
							endDefStrategiesCounterMax[i] = percentageAgentDidStrategy;
						} 
					}
				}
				//reset end strategies counter:
				endDefStrategiesCounterThisEvol = new int[NUM_STRATEGIES];
			}
		} 
		
	}
	
	
	public int getNumValidEvolutions() {
		return numValidEvolutions;
	}
	
}
